import logging

from collections import OrderedDict

from .signal import Signal, EpicsSignal, EpicsSignalRO
from .device import Device, Component as Cpt, DynamicDeviceComponent as DDC, Kind
from .areadetector import EpicsSignalWithRBV as SignalWithRBV


logger = logging.getLogger(__name__)


class ROI(Device):

    # 'name' is not an allowed attribute
    label = Cpt(EpicsSignal, "NM", lazy=True)
    count = Cpt(EpicsSignalRO, "", lazy=True)
    net_count = Cpt(EpicsSignalRO, "N", lazy=True)
    preset_count = Cpt(EpicsSignal, "P", lazy=True)
    is_preset = Cpt(EpicsSignal, "IP", lazy=True)
    bkgnd_chans = Cpt(EpicsSignal, "BG", lazy=True)
    hi_chan = Cpt(EpicsSignal, "HI", lazy=True)
    lo_chan = Cpt(EpicsSignal, "LO", lazy=True)

    def __init__(
        self,
        prefix,
        *,
        read_attrs=None,
        configuration_attrs=None,
        name=None,
        parent=None,
        **kwargs
    ):

        super().__init__(
            prefix,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            name=name,
            parent=parent,
            **kwargs
        )


def add_rois(range_, **kwargs):
    """Add one or more ROIs to an MCA instance

    Parameters
    ----------
    range_ : sequence of ints
        Must be be in the set [0,31]

    By default, an EpicsMCA is initialized with all 32 rois.
    These provide the following Components as EpicsSignals (N=[0,31]):
    EpicsMCA.rois.roiN.(label,count,net_count,preset_cnt, is_preset,
    bkgnd_chans, hi_chan, lo_chan)
    """
    defn = OrderedDict()

    for roi in range_:
        if not (0 <= roi < 32):
            raise ValueError("roi must be in the set [0,31]")

        attr = "roi{}".format(roi)
        defn[attr] = (ROI, ".R{}".format(roi), kwargs)

    return defn


class EpicsMCARecord(Device):
    """SynApps MCA Record interface"""

    stop_signal = Cpt(EpicsSignal, ".STOP", kind="omitted")
    preset_real_time = Cpt(EpicsSignal, ".PRTM", kind=Kind.config | Kind.normal)
    preset_live_time = Cpt(EpicsSignal, ".PLTM", kind="omitted")
    elapsed_real_time = Cpt(EpicsSignalRO, ".ERTM")
    elapsed_live_time = Cpt(EpicsSignalRO, ".ELTM", kind="omitted")

    spectrum = Cpt(EpicsSignalRO, ".VAL")
    background = Cpt(EpicsSignalRO, ".BG", kind="omitted")
    mode = Cpt(EpicsSignal, ".MODE", string=True, kind="omitted")

    rois = DDC(add_rois(range(0, 32), kind="omitted"), kind="omitted")

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        # could arguably be made a configuration_attr instead...
        self.stage_sigs["mode"] = "PHA"

    def stop(self, *, success=False):
        self.stop_signal.put(1)


class EpicsMCA(EpicsMCARecord):
    """mca records with extras from mca.db"""

    start = Cpt(EpicsSignal, "Start", kind="omitted")
    stop_signal = Cpt(EpicsSignal, "Stop", kind="omitted")
    erase = Cpt(EpicsSignal, "Erase", kind="omitted")
    erase_start = Cpt(EpicsSignal, "EraseStart", trigger_value=1, kind="omitted")

    check_acquiring = Cpt(EpicsSignal, "CheckACQG", kind="omitted")
    client_wait = Cpt(EpicsSignal, "ClientWait", kind="omitted")
    enable_wait = Cpt(EpicsSignal, "EnableWait", kind="omitted")
    force_read = Cpt(EpicsSignal, "Read", kind="omitted")
    set_client_wait = Cpt(EpicsSignal, "SetClientWait", kind="omitted")
    status = Cpt(EpicsSignal, "Status", kind="omitted")
    when_acq_stops = Cpt(EpicsSignal, "WhenAcqStops", kind="omitted")
    why1 = Cpt(EpicsSignal, "Why1", kind="omitted")
    why2 = Cpt(EpicsSignal, "Why2", kind="omitted")
    why3 = Cpt(EpicsSignal, "Why3", kind="omitted")
    why4 = Cpt(EpicsSignal, "Why4", kind="omitted")


class EpicsMCAReadNotify(EpicsMCARecord):
    """mca record with extras from mcaReadNotify.db"""

    start = Cpt(EpicsSignal, "Start", kind="omitted")
    stop_signal = Cpt(EpicsSignal, "Stop", kind="omitted")
    erase = Cpt(EpicsSignal, "Erase", kind="omitted")
    erase_start = Cpt(EpicsSignal, "EraseStart", trigger_value=1, kind="omitted")

    check_acquiring = Cpt(EpicsSignal, "CheckACQG", kind="omitted")
    client_wait = Cpt(EpicsSignal, "ClientWait", kind="omitted")
    enable_wait = Cpt(EpicsSignal, "EnableWait", kind="omitted")
    force_read = Cpt(EpicsSignal, "Read", kind="omitted")
    set_client_wait = Cpt(EpicsSignal, "SetClientWait", kind="omitted")
    status = Cpt(EpicsSignal, "Status", kind="omitted")


class EpicsMCACallback(Device):
    """Callback-related signals for MCA devices"""

    read_callback = Cpt(EpicsSignal, "ReadCallback")
    read_data_once = Cpt(EpicsSignal, "ReadDataOnce")
    read_status_once = Cpt(EpicsSignal, "ReadStatusOnce")
    collect_data = Cpt(EpicsSignal, "CollectData")


class EpicsDXP(Device):
    """All high-level DXP parameters for each channel"""

    preset_mode = Cpt(EpicsSignal, "PresetMode", string=True)

    live_time_output = Cpt(SignalWithRBV, "LiveTimeOutput", string=True)
    elapsed_live_time = Cpt(EpicsSignal, "ElapsedLiveTime")
    elapsed_real_time = Cpt(EpicsSignal, "ElapsedRealTime")
    elapsed_trigger_live_time = Cpt(EpicsSignal, "ElapsedTriggerLiveTime")

    # Trigger Filter PVs
    trigger_peaking_time = Cpt(SignalWithRBV, "TriggerPeakingTime")
    trigger_threshold = Cpt(SignalWithRBV, "TriggerThreshold")
    trigger_gap_time = Cpt(SignalWithRBV, "TriggerGapTime")
    trigger_output = Cpt(SignalWithRBV, "TriggerOutput", string=True)
    max_width = Cpt(SignalWithRBV, "MaxWidth")

    # Energy Filter PVs
    peaking_time = Cpt(SignalWithRBV, "PeakingTime")
    energy_threshold = Cpt(SignalWithRBV, "EnergyThreshold")
    gap_time = Cpt(SignalWithRBV, "GapTime")

    # Baseline PVs
    baseline_cut_percent = Cpt(SignalWithRBV, "BaselineCutPercent")
    baseline_cut_enable = Cpt(SignalWithRBV, "BaselineCutEnable")
    baseline_filter_length = Cpt(SignalWithRBV, "BaselineFilterLength")
    baseline_threshold = Cpt(SignalWithRBV, "BaselineThreshold")
    baseline_energy_array = Cpt(EpicsSignal, "BaselineEnergyArray")
    baseline_histogram = Cpt(EpicsSignal, "BaselineHistogram")
    baseline_threshold = Cpt(SignalWithRBV, "BaselineThreshold")

    # Misc PVs
    preamp_gain = Cpt(SignalWithRBV, "PreampGain")
    detector_polarity = Cpt(SignalWithRBV, "DetectorPolarity")
    reset_delay = Cpt(SignalWithRBV, "ResetDelay")
    decay_time = Cpt(SignalWithRBV, "DecayTime")
    max_energy = Cpt(SignalWithRBV, "MaxEnergy")
    adc_percent_rule = Cpt(SignalWithRBV, "ADCPercentRule")
    max_width = Cpt(SignalWithRBV, "MaxWidth")

    # read-only diagnostics
    triggers = Cpt(EpicsSignalRO, "Triggers", lazy=True)
    events = Cpt(EpicsSignalRO, "Events", lazy=True)
    overflows = Cpt(EpicsSignalRO, "Overflows", lazy=True)
    underflows = Cpt(EpicsSignalRO, "Underflows", lazy=True)
    input_count_rate = Cpt(EpicsSignalRO, "InputCountRate", lazy=True)
    output_count_rate = Cpt(EpicsSignalRO, "OutputCountRate", lazy=True)

    mca_bin_width = Cpt(EpicsSignalRO, "MCABinWidth_RBV")
    calibration_energy = Cpt(EpicsSignalRO, "CalibrationEnergy_RBV")
    current_pixel = Cpt(EpicsSignal, "CurrentPixel")
    dynamic_range = Cpt(EpicsSignalRO, "DynamicRange_RBV")

    # Preset options
    preset_events = Cpt(SignalWithRBV, "PresetEvents")
    preset_mode = Cpt(SignalWithRBV, "PresetMode", string=True)
    preset_triggers = Cpt(SignalWithRBV, "PresetTriggers")

    # Trace options
    trace_data = Cpt(EpicsSignal, "TraceData")
    trace_mode = Cpt(SignalWithRBV, "TraceMode", string=True)
    trace_time_array = Cpt(EpicsSignal, "TraceTimeArray")
    trace_time = Cpt(SignalWithRBV, "TraceTime")


class EpicsDXPLowLevelParameter(Device):
    param_name = Cpt(EpicsSignal, "Name")
    value = Cpt(SignalWithRBV, "Val")


class EpicsDXPLowLevel(Device):
    num_low_level_params = Cpt(EpicsSignal, "NumLLParams")
    read_low_level_params = Cpt(EpicsSignal, "ReadLLParams")

    parameter_prefix = "LL{}"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._parameter_cache = {}

    def get_low_level_parameter(self, index):
        """Get a DXP low level parameter

        Parameters
        ----------
        index : int
            In the range of [0, 229]

        Returns
        -------
        param : EpicsDXPLowLevelParameter
        """
        try:
            return self._parameter_cache[index]
        except KeyError:
            pass

        prefix = "{}{}".format(self.prefix, self.parameter_prefix)
        name = "{}_param{}".format(self.name, index)
        param = EpicsDXPLowLevelParameter(prefix, name=name)
        self._parameter_cache[index] = param
        return param


class EpicsDXPMapping(Device):
    apply = Cpt(EpicsSignal, "Apply")
    auto_apply = Cpt(SignalWithRBV, "AutoApply")
    auto_pixels_per_buffer = Cpt(SignalWithRBV, "AutoPixelsPerBuffer")
    buffer_size = Cpt(EpicsSignalRO, "BufferSize_RBV")
    collect_mode = Cpt(SignalWithRBV, "CollectMode")
    ignore_gate = Cpt(SignalWithRBV, "IgnoreGate")
    input_logic_polarity = Cpt(SignalWithRBV, "InputLogicPolarity")
    list_mode = Cpt(SignalWithRBV, "ListMode")
    mbytes_read = Cpt(EpicsSignalRO, "MBytesRead_RBV")
    next_pixel = Cpt(EpicsSignal, "NextPixel")
    pixel_advance_mode = Cpt(SignalWithRBV, "PixelAdvanceMode")
    pixels_per_buffer = Cpt(SignalWithRBV, "PixelsPerBuffer")
    pixels_per_run = Cpt(SignalWithRBV, "PixelsPerRun")
    read_rate = Cpt(EpicsSignalRO, "ReadRate_RBV")
    sync_count = Cpt(SignalWithRBV, "SyncCount")


class EpicsDXPBaseSystem(Device):
    channel_advance = Cpt(EpicsSignal, "ChannelAdvance")
    client_wait = Cpt(EpicsSignal, "ClientWait")
    dwell = Cpt(EpicsSignal, "Dwell")
    max_scas = Cpt(EpicsSignal, "MaxSCAs")
    num_scas = Cpt(SignalWithRBV, "NumSCAs")
    poll_time = Cpt(SignalWithRBV, "PollTime")
    prescale = Cpt(EpicsSignal, "Prescale")
    save_system = Cpt(SignalWithRBV, "SaveSystem")
    save_system_file = Cpt(EpicsSignal, "SaveSystemFile")
    set_client_wait = Cpt(EpicsSignal, "SetClientWait")


class EpicsDXPMultiElementSystem(EpicsDXPBaseSystem):
    # Preset info
    preset_events = Cpt(EpicsSignal, "PresetEvents")
    preset_live_time = Cpt(EpicsSignal, "PresetLive")
    preset_real_time = Cpt(EpicsSignal, "PresetReal")
    preset_mode = Cpt(EpicsSignal, "PresetMode", string=True)
    preset_triggers = Cpt(EpicsSignal, "PresetTriggers")

    # Acquisition
    erase_all = Cpt(EpicsSignal, "EraseAll")
    erase_start = Cpt(EpicsSignal, "EraseStart", trigger_value=1)
    start_all = Cpt(EpicsSignal, "StartAll")
    stop_all = Cpt(EpicsSignal, "StopAll")

    # Status
    set_acquire_busy = Cpt(EpicsSignal, "SetAcquireBusy")
    acquire_busy = Cpt(EpicsSignal, "AcquireBusy")
    status_all = Cpt(EpicsSignal, "StatusAll")
    status_all_once = Cpt(EpicsSignal, "StatusAllOnce")
    acquiring = Cpt(EpicsSignal, "Acquiring")

    # Reading
    read_baseline_histograms = Cpt(EpicsSignal, "ReadBaselineHistograms")
    read_all = Cpt(EpicsSignal, "ReadAll")
    read_all_once = Cpt(EpicsSignal, "ReadAllOnce")

    # As a debugging note, if snl_connected is not '1', your IOC is
    # misconfigured:
    snl_connected = Cpt(EpicsSignal, "SNL_Connected")

    # Copying to individual elements
    copy_adcp_ercent_rule = Cpt(EpicsSignal, "CopyADCPercentRule")
    copy_baseline_cut_enable = Cpt(EpicsSignal, "CopyBaselineCutEnable")
    copy_baseline_cut_percent = Cpt(EpicsSignal, "CopyBaselineCutPercent")
    copy_baseline_filter_length = Cpt(EpicsSignal, "CopyBaselineFilterLength")
    copy_baseline_threshold = Cpt(EpicsSignal, "CopyBaselineThreshold")
    copy_decay_time = Cpt(EpicsSignal, "CopyDecayTime")
    copy_detector_polarity = Cpt(EpicsSignal, "CopyDetectorPolarity")
    copy_energy_threshold = Cpt(EpicsSignal, "CopyEnergyThreshold")
    copy_gap_time = Cpt(EpicsSignal, "CopyGapTime")
    copy_max_energy = Cpt(EpicsSignal, "CopyMaxEnergy")
    copy_max_width = Cpt(EpicsSignal, "CopyMaxWidth")
    copy_peaking_time = Cpt(EpicsSignal, "CopyPeakingTime")
    copy_preamp_gain = Cpt(EpicsSignal, "CopyPreampGain")
    copy_roic_hannel = Cpt(EpicsSignal, "CopyROIChannel")
    copy_roie_nergy = Cpt(EpicsSignal, "CopyROIEnergy")
    copy_roi_sca = Cpt(EpicsSignal, "CopyROI_SCA")
    copy_reset_delay = Cpt(EpicsSignal, "CopyResetDelay")
    copy_trigger_gap_time = Cpt(EpicsSignal, "CopyTriggerGapTime")
    copy_trigger_peaking_time = Cpt(EpicsSignal, "CopyTriggerPeakingTime")
    copy_trigger_threshold = Cpt(EpicsSignal, "CopyTriggerThreshold")

    # do_* executes the process:
    do_read_all = Cpt(EpicsSignal, "DoReadAll")
    do_read_baseline_histograms = Cpt(EpicsSignal, "DoReadBaselineHistograms")
    do_read_traces = Cpt(EpicsSignal, "DoReadTraces")
    do_status_all = Cpt(EpicsSignal, "DoStatusAll")

    # Time
    dead_time = Cpt(EpicsSignal, "DeadTime")
    elapsed_live = Cpt(EpicsSignal, "ElapsedLive")
    elapsed_real = Cpt(EpicsSignal, "ElapsedReal")
    idead_time = Cpt(EpicsSignal, "IDeadTime")

    # low-level
    read_low_level_params = Cpt(EpicsSignal, "ReadLLParams")

    # Traces
    read_traces = Cpt(EpicsSignal, "ReadTraces")
    trace_modes = Cpt(EpicsSignal, "TraceModes", string=True)
    trace_times = Cpt(EpicsSignal, "TraceTimes")


class SaturnMCA(EpicsMCA, EpicsMCACallback):
    pass


class SaturnDXP(EpicsDXP, EpicsDXPLowLevel):
    pass


class Saturn(EpicsDXPBaseSystem):
    """DXP Saturn with 1 channel example"""

    dxp = Cpt(SaturnDXP, "dxp1:")
    mca = Cpt(SaturnMCA, "mca1")


class MercuryDXP(EpicsDXP, EpicsDXPLowLevel):
    pass


class Mercury1(EpicsDXPMultiElementSystem):
    """DXP Mercury with 1 channel example"""

    dxp = Cpt(MercuryDXP, "dxp1:")
    mca = Cpt(EpicsMCARecord, "mca1")


class SoftDXPTrigger(Device):
    """Simple soft trigger for DXP devices

    Parameters
    ----------
    count_signal : str, optional
        Signal to set acquisition time (default: 'preset_real_time')
    preset_mode : str, optional
        Default preset mode for the stage signals (default: 'Real time')
    mode_signal : str, optional
        Preset mode signal attribute (default 'preset_mode')
    stop_signal : str, optional
        Stop signal attribute (default 'stop_all')
    """

    count_time = Cpt(Signal, value=None, doc="bluesky count time")

    def __init__(
        self,
        *args,
        count_signal="preset_real_time",
        stop_signal="stop_all",
        mode_signal="preset_mode",
        preset_mode="Real time",
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._status = None
        self._count_signal = getattr(self, count_signal)

        stop_signal = getattr(self, stop_signal)
        self.stage_sigs[stop_signal] = 1

        mode_signal = getattr(self, mode_signal)
        self.stage_sigs[mode_signal] = preset_mode

    def stage(self):
        if self.count_time.get() is None:
            # remove count_time from the stage signals if count_time unset
            try:
                del self.stage_sigs[self._count_signal]
            except KeyError:
                pass
        else:
            self.stage_sigs[self._count_signal] = self.count_time.get()

        super().stage()



import functools
import time
import weakref
from enum import IntFlag
from itertools import count
from logging import LoggerAdapter, getLogger
from typing import ClassVar, FrozenSet

from .log import control_layer_logger


def select_version(cls, version):
    """Select closest compatible version to requested version

    Compatible is defined as ``class_version <= requested_version``
    as defined by the types used to denote the versions.

    Parameters
    ----------
    cls : type
        The base class to find a version of

    version : any
        Must be the same type as used to define the class versions.

    """
    all_versions = cls._class_info_["versions"]
    matched_version = max(ver for ver in all_versions if ver <= version)
    return all_versions[matched_version]


try:
    from enum import KEEP

    class IFBase(IntFlag, boundary=KEEP):
        ...

except ImportError:

    IFBase = IntFlag


class Kind(IFBase):
    """
    This is used in the .kind attribute of all OphydObj (Signals, Devices).

    A Device examines its components' .kind atttribute to decide whether to
    traverse it in read(), read_configuration(), or neither. Additionally, if
    decides whether to include its name in `hints['fields']`.
    """

    omitted = 0b000
    normal = 0b001
    config = 0b010
    hinted = 0b101  # Notice that bool(hinted & normal) is True.


class UnknownSubscription(KeyError):
    "Subclass of KeyError.  Raised for unknown event type"
    ...


def register_instances_keyed_on_name(fail_if_late=False):
    """Register OphydObj instances in a WeakValueDictionary keyed on name.

    Be advised that ophyd does not require 'name' to be unique and is
    configurable by the user at run-time so this should
    not be relied on unless name uniqueness is enforced by other means.

    Parameters
    ----------
    fail_if_late : boolean
        If True, verify that OphydObj has not yet been instantiated and raise
        ``RuntimeError`` if it has, as a way of verify that no instances will
        be "missed" by this registry. False by default.

    Returns
    -------
    WeakValueDictionary
    """
    weak_dict = weakref.WeakValueDictionary()

    def register(instance):
        weak_dict[instance.name] = instance

    OphydObject.add_instantiation_callback(register, fail_if_late)
    return weak_dict


def register_instances_in_weakset(fail_if_late=False):
    """Register OphydObj instances in a WeakSet.

    Be advised that OphydObj may not always be hashable.

    Parameters
    ----------
    fail_if_late : boolean
        If True, verify that OphydObj has not yet been instantiated and raise
        ``RuntimeError`` if it has, as a way of verify that no instances will
        be "missed" by this registry. False by default.

    Returns
    -------
    WeakSet
    """
    weak_set = weakref.WeakSet()

    def register(instance):
        weak_set.add(instance)

    OphydObject.add_instantiation_callback(register, fail_if_late)
    return weak_set


class OphydObject:
    """The base class for all objects in Ophyd

    Handles:

      * Subscription/callback mechanism

    Parameters
    ----------
    name : str, optional
        The name of the object.
    attr_name : str, optional
        The attr name on it's parent (if it has one)
        ex ``getattr(self.parent, self.attr_name) is self``
    parent : parent, optional
        The object's parent, if it exists in a hierarchy
    kind : a member of the :class:`~ophydobj.Kind` :class:`~enum.IntEnum`
        (or equivalent integer), optional
        Default is ``Kind.normal``. See :class:`~ophydobj.Kind` for options.

    Attributes
    ----------
    name
    """

    # Any callables appended to this mutable class variable will be notified
    # one time when a new instance of OphydObj is instantiated. See
    # OphydObject.add_instantiation_callback().
    __instantiation_callbacks = []
    _default_sub = None
    # This is set to True when the first OphydObj is instantiated. This may be
    # of interest to code that adds something to instantiation_callbacks, which
    # may want to know whether it has already "missed" any instances.
    __any_instantiated = False
    subscriptions: ClassVar[FrozenSet[str]] = frozenset()

    def __init__(self, *, name=None, attr_name="", parent=None, labels=None, kind=None):
        if labels is None:
            labels = set()
        self._ophyd_labels_ = set(labels)
        if kind is None:
            kind = Kind.normal
        self.kind = kind

        super().__init__()

        # base name and ref to parent, these go with properties
        if name is None:
            name = ""
        self._attr_name = attr_name
        if not isinstance(name, str):
            raise ValueError("name must be a string.")
        self._name = name
        self._parent = parent

        # dictionary of wrapped callbacks
        self._callbacks = {k: {} for k in self.subscriptions}
        # this is to maintain api on clear_sub
        self._unwrapped_callbacks = {k: {} for k in self.subscriptions}
        # map cid -> back to which event it is in
        self._cid_to_event_mapping = dict()
        # cache of last inputs to _run_subs, the semi-private way
        # to trigger the callbacks for a given subscription to be run
        self._args_cache = {k: None for k in self.subscriptions}
        # count of subscriptions we have handed out, used to give unique ids
        self._cb_count = count()
        self.log = LoggerAdapter(
            getLogger("ophyd.objects"), {"ophyd_object_name": name}
        )
        self.control_layer_log = LoggerAdapter(
            control_layer_logger, {"ophyd_object_name": name}
        )

        if not self.__any_instantiated:
            self.log.debug("first instance of OphydObject: id=%s", id(self))
            OphydObject._mark_as_instantiated()
        self.__register_instance(self)

    @classmethod
    def _mark_as_instantiated(cls):
        cls.__any_instantiated = True

    @classmethod
    def add_instantiation_callback(cls, callback, fail_if_late=False):
        """
        Register a callback which will receive each OphydObject instance.

        Parameters
        ----------
        callback : callable
            Expected signature: ``f(ophydobj_instance)``
        fail_if_late : boolean
            If True, verify that OphydObj has not yet been instantiated and raise
            ``RuntimeError`` if it has, as a way of verify that no instances will
            be "missed" by this registry. False by default.
        """
        if fail_if_late and OphydObject.__any_instantiated:
            raise RuntimeError(
                "OphydObject has already been instantiated at least once, and "
                "this callback will not be notified of those instances that "
                "have already been created. If that is acceptable for this "
                "application, set fail_if_false=False."
            )
        # This is a class variable.
        cls.__instantiation_callbacks.append(callback)

    @classmethod
    def __register_instance(cls, instance):
        """
        Notify the callbacks in OphydObject.instantiation_callbacks of an instance.
        """
        for callback in cls.__instantiation_callbacks:
            callback(instance)

    def __init_subclass__(
        cls, version=None, version_of=None, version_type=None, **kwargs
    ):
        "This is called automatically in Python for all subclasses of OphydObject"
        super().__init_subclass__(**kwargs)

        cls.subscriptions = frozenset(
            {
                getattr(cls, key)
                for key in dir(cls)
                if key.startswith("SUB") or key.startswith("_SUB")
            }
        )

        if version is None:
            if version_of is not None:
                raise RuntimeError(
                    "Must specify a version if `version_of` " "is specified"
                )
            if version_type is None:
                return
            # Allow specification of version_type without specifying a version,
            # for use in a base class

            cls._class_info_ = dict(
                versions={},
                version=None,
                version_type=version_type,
                version_of=version_of,
            )
            return

        if version_of is None:
            versions = {}
            version_of = cls
        else:
            versions = version_of._class_info_["versions"]
            if version_type is None:
                version_type = version_of._class_info_["version_type"]

            elif version_type != version_of._class_info_["version_type"]:
                raise RuntimeError(
                    "version_type with in a family must be consistent, "
                    f"you passed in {version_type}, to {cls.__name__} "
                    f"but {version_of.__name__} has version_type "
                    f"{version_of._class_info_['version_type']}"
                )

            if not issubclass(cls, version_of):
                raise RuntimeError(
                    f"Versions are only valid for classes in the same "
                    f"hierarchy. {cls.__name__} is not a subclass of "
                    f"{version_of.__name__}."
                )

        if versions is not None and version in versions:
            getLogger("ophyd.object").warning(
                "Redefining %r version %s: old=%r new=%r",
                version_of,
                version,
                versions[version],
                cls,
            )

        versions[version] = cls

        cls._class_info_ = dict(
            versions=versions,
            version=version,
            version_type=version_type,
            version_of=version_of,
        )

    def _validate_kind(self, val):
        if isinstance(val, str):
            return Kind[val.lower()]
        return Kind(val)

    @property
    def kind(self):
        return self._kind

    @kind.setter
    def kind(self, val):
        self._kind = self._validate_kind(val)

    @property
    def dotted_name(self) -> str:
        """Return the dotted name"""
        names = []
        obj = self
        while obj.parent is not None:
            names.append(obj.attr_name)
            obj = obj.parent
        return ".".join(names[::-1])

    @property
    def name(self):
        """name of the device"""
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def attr_name(self):
        return self._attr_name

    @property
    def connected(self):
        """If the device is connected.

        Subclasses should override this"""
        return True

    def destroy(self):
        """Disconnect the object from the underlying control layer"""
        self.unsubscribe_all()

    @property
    def parent(self):
        """The parent of the ophyd object.

        If at the top of its hierarchy, `parent` will be None
        """
        return self._parent

    @property
    def root(self):
        "Walk parents to find ultimate ancestor (parent's parent...)."
        root = self
        while True:
            if root.parent is None:
                return root
            root = root.parent

    @property
    def report(self):
        """A report on the object."""
        return {}

    @property
    def event_types(self):
        """Events that can be subscribed to via `obj.subscribe`"""
        return tuple(self.subscriptions)

    def _run_subs(self, *args, sub_type, **kwargs):
        """Run a set of subscription callbacks

        Only the kwarg ``sub_type`` is required, indicating
        the type of callback to perform. All other positional arguments
        and kwargs are passed directly to the callback function.

        The host object will be injected into kwargs as 'obj' unless that key
        already exists.

        If the `timestamp` is None, then it will be replaced by the current
        time.

        No exceptions are raised if the callback functions fail.
        """
        if sub_type not in self.subscriptions:
            raise UnknownSubscription(
                "Unknown subscription {!r}, must be one of {!r}".format(
                    sub_type, self.subscriptions
                )
            )

        kwargs["sub_type"] = sub_type
        # Guarantee that the object will be in the kwargs
        kwargs.setdefault("obj", self)

        # And if a timestamp key exists, but isn't filled -- supply it with
        # a new timestamp
        if "timestamp" in kwargs and kwargs["timestamp"] is None:
            kwargs["timestamp"] = time.time()

        # Shallow-copy the callback arguments for replaying the
        # callback at a later time (e.g., when a new subscription is made)
        self._args_cache[sub_type] = (tuple(args), dict(kwargs))

        for cb in list(self._callbacks[sub_type].values()):
            cb(*args, **kwargs)

    def subscribe(self, callback, event_type=None, run=True):
        """Subscribe to events this event_type generates.

        The callback will be called as ``cb(*args, **kwargs)`` with
        the values passed to `_run_subs` with the following additional keys:

           sub_type : the string value of the event_type
           obj : the host object, added if 'obj' not already in kwargs

        if the key 'timestamp' is in kwargs _and_ is None, then it will
        be replaced with the current time before running the callback.

        The ``*args``, ``**kwargs`` passed to _run_subs will be cached as
        shallow copies, be aware of passing in mutable data.

        .. warning::

           If the callback raises any exceptions when run they will be
           silently ignored.

        Parameters
        ----------
        callback : callable
            A callable function (that takes kwargs) to be run when the event is
            generated.  The expected signature is ::

              def cb(*args, obj: OphydObject, sub_type: str, **kwargs) -> None:

            The exact args/kwargs passed are whatever are passed to
            ``_run_subs``
        event_type : str, optional
            The name of the event to subscribe to (if None, defaults to
            the default sub for the instance - obj._default_sub)

            This maps to the ``sub_type`` kwargs in `_run_subs`
        run : bool, optional
            Run the callback now

        See Also
        --------
        clear_sub, _run_subs

        Returns
        -------
        cid : int
            id of callback, can be passed to `unsubscribe` to remove the
            callback

        """
        if not callable(callback):
            raise ValueError("callback must be callable")
        # do default event type
        if event_type is None:
            # warnings.warn("Please specify which call back you wish to "
            #               "attach to defaulting to {}"
            #               .format(self._default_sub), stacklevel=2)
            event_type = self._default_sub

        if event_type is None:
            raise ValueError(
                "Subscription type not set and object {} of class"
                " {} has no default subscription set"
                "".format(self.name, self.__class__.__name__)
            )

        # check that this is a valid event type
        if event_type not in self.subscriptions:
            raise UnknownSubscription(
                "Unknown subscription {!r}, must be one of {!r}".format(
                    event_type, self.subscriptions
                )
            )

        # wrapper for callback to snarf exceptions
        def wrap_cb(cb):
            @functools.wraps(cb)
            def inner(*args, **kwargs):
                try:
                    cb(*args, **kwargs)
                except Exception:
                    sub_type = kwargs["sub_type"]
                    self.log.exception(
                        "Subscription %s callback exception (%s)", sub_type, self
                    )

            return inner

        # get next cid
        cid = next(self._cb_count)
        wrapped = wrap_cb(callback)
        self._unwrapped_callbacks[event_type][cid] = callback
        self._callbacks[event_type][cid] = wrapped
        self._cid_to_event_mapping[cid] = event_type

        if run:
            cached = self._args_cache[event_type]
            if cached is not None:
                args, kwargs = cached
                wrapped(*args, **kwargs)

        return cid

    def _reset_sub(self, event_type):
        """Remove all subscriptions in an event type"""
        self._callbacks[event_type].clear()
        self._unwrapped_callbacks[event_type].clear()

    def clear_sub(self, cb, event_type=None):
        """Remove a subscription, given the original callback function

        See also :meth:`subscribe`, :meth:`unsubscribe`

        Parameters
        ----------
        cb : callable
            The callback
        event_type : str, optional
            The event to unsubscribe from (if None, removes it from all event
            types)
        """
        if event_type is None:
            event_types = self.event_types
        else:
            event_types = [event_type]
        cid_list = []
        for et in event_types:
            for cid, target in self._unwrapped_callbacks[et].items():
                if cb == target:
                    cid_list.append(cid)
        for cid in cid_list:
            self.unsubscribe(cid)

    def unsubscribe(self, cid):
        """Remove a subscription

        See also :meth:`subscribe`, :meth:`clear_sub`

        Parameters
        ----------
        cid : int
           token return by :meth:`subscribe`
        """
        ev_type = self._cid_to_event_mapping.pop(cid, None)
        if ev_type is None:
            return
        del self._unwrapped_callbacks[ev_type][cid]
        del self._callbacks[ev_type][cid]

    def unsubscribe_all(self):
        for ev_type in self._callbacks:
            self._reset_sub(ev_type)

    def check_value(self, value, **kwargs):
        """Check if the value is valid for this object

        This function does no normalization, but may raise if the
        value is invalid.

        Raises
        ------
        ValueError
        """
        pass

    def __repr__(self):
        info = self._repr_info()
        info = ", ".join("{}={!r}".format(key, value) for key, value in info)
        return "{}({})".format(self.__class__.__name__, info)

    def _repr_info(self):
        "Yields pairs of (key, value) to generate the object repr"
        if self.name is not None:
            yield ("name", self.name)

        if self._parent is not None:
            yield ("parent", self.parent.name)

    def __copy__(self):
        """Copy the ophyd object

        Shallow copying ophyd objects uses the repr information from the
        _repr_info method to create a new object.
        """
        kwargs = dict(self._repr_info())
        return self.__class__(**kwargs)

    def __getnewargs_ex__(self):
        """Used by pickle to serialize an ophyd object

        Returns
        -------
        (args, kwargs)
            Arguments to be passed to __init__, necessary to recreate this
            object
        """
        kwargs = dict(self._repr_info())
        return ((), kwargs)



import logging
import time
from collections import OrderedDict
import functools
from .ophydobj import OphydObject, Kind
from .status import MoveStatus, wait as status_wait, StatusBase
from .utils.epics_pvs import data_type, data_shape
from .utils.errors import LimitError
from typing import Any, Callable


logger = logging.getLogger(__name__)


class PositionerBase(OphydObject):
    """The positioner base class

    Subclass from this to implement your own positioners.

    .. note ::

       Subclasses should add an additional 'wait' keyword argument on
       the move method. The `MoveStatus` object returned from
       `PositionerBase` can then be waited on after the subclass
       finishes the motion configuration.

    """

    SUB_START = "start_moving"
    SUB_DONE = "done_moving"
    SUB_READBACK = "readback"
    _SUB_REQ_DONE = "_req_done"  # requested move finished subscription
    _default_sub = SUB_READBACK

    def __init__(
        self, *, name=None, parent=None, settle_time=0.0, timeout=None, **kwargs
    ):
        super().__init__(name=name, parent=parent, **kwargs)

        self._started_moving = False
        self._moving = False
        self._position = None
        self._settle_time = settle_time
        self._timeout = timeout

    # High level
    def set(
        self,
        new_position: Any,
        *,
        timeout: float = None,
        moved_cb: Callable = None,
        wait: bool = False,
    ) -> StatusBase:
        """Set a value and return a Status object


        Parameters
        ----------
        new_position : object

            The input here is whatever the device requires (this
            should be over-ridden by the implementation.  For example
            a motor would take a float, a shutter the strings {'Open',
            'Close'}, and a goineometer (h, k, l) tuples

        timeout : float, optional

            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        moved_cb : callable, optional
            Deprecated

            Call this callback when movement has finished. This callback
            must accept one keyword argument: 'obj' which will be set to
            this positioner instance.

        wait : bool, optional
            Deprecated

            If the method should block until the Status object reports
            it is done.

            Defaults to False

        Returns
        -------
        status : StatusBase
            Status object to indicate when the motion / set is done.
        """
        return self.move(new_position, wait=wait, moved_cb=moved_cb, timeout=timeout)

    def stop(self, *, success: bool = False):
        """Stops motion.

        Sub-classes must extend this method to _actually_ stop the device.

        Parameters
        ----------
        success : bool, optional
            If the move should be considered a success despite the stop.

            Defaults to False
        """
        self._done_moving(success=success)

    # Suggested properties
    @property
    def settle_time(self):
        """Amount of time to wait after moves to report status completion"""
        return self._settle_time

    @settle_time.setter
    def settle_time(self, settle_time):
        self._settle_time = settle_time

    @property
    def timeout(self):
        """Amount of time to wait before to considering a motion as failed"""
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        if timeout is None:
            self._timeout = None
        else:
            self._timeout = float(timeout)

    # low level
    @property
    def report(self):
        rep = super().report
        rep["position"] = self.position
        return rep

    @property
    def egu(self):
        """The engineering units (EGU) for positions"""
        raise NotImplementedError("Subclass must implement egu")

    @property
    def limits(self):
        return (0, 0)

    @property
    def low_limit(self):
        return self.limits[0]

    @property
    def high_limit(self):
        return self.limits[1]

    def move(self, position, moved_cb=None, timeout=None):
        """Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback
            must accept one keyword argument: 'obj' which will be set to
            this positioner instance. It should also accept a positional
            argument 'status', the Status object.

            .. versionchanged:: 1.5.0

               The expected signature changed from ``moved_cb(*, obj)`` to
               ``moved_cb(status, *, obj)``. The old signature is still
               supported, but a warning will be issued.

        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        """

        if timeout is None:
            timeout = self._timeout

        self.check_value(position)

        self._run_subs(sub_type=self._SUB_REQ_DONE, success=False)
        self._reset_sub(self._SUB_REQ_DONE)

        status = MoveStatus(
            self, position, timeout=timeout, settle_time=self._settle_time
        )

        if moved_cb is not None:
            status.add_callback(functools.partial(moved_cb, obj=self))
            # the status object will run this callback when finished

        self.subscribe(status._finished, event_type=self._SUB_REQ_DONE, run=False)

        return status

    def _done_moving(self, success=True, timestamp=None, value=None, **kwargs):
        """Call when motion has completed.  Runs ``SUB_DONE`` subscription."""
        if success:
            self._run_subs(sub_type=self.SUB_DONE, timestamp=timestamp, value=value)

        self._run_subs(
            sub_type=self._SUB_REQ_DONE, success=success, timestamp=timestamp
        )
        self._reset_sub(self._SUB_REQ_DONE)

    @property
    def position(self):
        """The current position of the motor in its engineering units

        Returns
        -------
        position : any
        """
        return self._position

    def _set_position(self, value, **kwargs):
        """Set the current internal position, run the readback subscription"""
        self._position = value

        timestamp = kwargs.pop("timestamp", time.time())
        self._run_subs(
            sub_type=self.SUB_READBACK, timestamp=timestamp, value=value, **kwargs
        )

    @property
    def moving(self):
        """Whether or not the motor is moving

        Returns
        -------
        moving : bool
        """
        return self._moving

    def _repr_info(self):
        yield from super()._repr_info()
        yield ("settle_time", self._settle_time)
        yield ("timeout", self._timeout)

    @property
    def hints(self):
        if (~Kind.normal & Kind.hinted) & self.kind:
            return {"fields": [self.name]}
        else:
            return {"fields": []}


class SoftPositioner(PositionerBase):
    """A positioner which does not communicate with any hardware

    SoftPositioner 'moves' immediately to the target position when commanded to
    do so.

    Parameters
    ----------
    limits : (low_limit, high_limit)
        Soft limits to use
    egu : str, optional
        Engineering units (EGU) for a position
    source : str, optional
        Metadata indicating the source of this positioner's position. Defaults
        to 'computed'
    init_pos : float, optional
        Create the positioner with this starting position.  Defaults to ``None``.
    """

    def __init__(
        self, *, egu="", limits=None, source="computed", init_pos=None, **kwargs
    ):
        super().__init__(**kwargs)

        self._egu = egu
        if limits is None:
            limits = (0, 0)

        self._limits = tuple(limits)
        self.source = source
        if init_pos is not None:
            self.set(init_pos)

    @property
    def limits(self):
        return self._limits

    @property
    def egu(self):
        """The engineering units (EGU) for positions"""
        return self._egu

    def _setup_move(self, position, status):
        """Move requested to position

        This is a SoftPositioner method which allows customization of what
        happens when a motion request happens without re-implementing
        all of `move`.

        Parameters
        ----------
        position : any
            Position to move to (already verified by `check_value`)
        status : MoveStatus
            Status object created by PositionerBase.move()
        """
        # A soft positioner immediately 'moves' to the target position when
        # requested.
        self._run_subs(sub_type=self.SUB_START, timestamp=time.time())

        self._started_moving = True
        self._moving = False

        self._set_position(position)
        self._done_moving()

    def move(self, position, wait=True, timeout=None, moved_cb=None):
        """Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback
            must accept one keyword argument: 'obj' which will be set to
            this positioner instance.
        wait : bool, optional
            Wait until motion has completed
        timeout : float, optional
            Maximum time to wait for a motion

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        """
        status = super().move(position, moved_cb=moved_cb, timeout=timeout)

        self._setup_move(position, status)

        if wait:
            try:
                status_wait(status)
            except RuntimeError:
                raise RuntimeError("Motion did not complete successfully")

        return status

    def _repr_info(self):
        yield from super()._repr_info()
        yield ("egu", self._egu)
        yield ("limits", self._limits)
        yield ("source", self.source)

    def read(self):
        d = OrderedDict()
        d[self.name] = {"value": self.position, "timestamp": time.time()}
        return d

    def describe(self):
        """Return the description as a dictionary

        Returns
        -------
        dict
            Dictionary of name and formatted description string
        """
        desc = OrderedDict()
        desc[self.name] = {
            "source": str(self.source),
            "dtype": data_type(self.position),
            "shape": data_shape(self.position),
            "units": self.egu,
            "lower_ctrl_limit": self.low_limit,
            "upper_ctrl_limit": self.high_limit,
        }
        return desc

    def read_configuration(self):
        return OrderedDict()

    def describe_configuration(self):
        return OrderedDict()

    def check_value(self, pos):
        """Check that the position is within the soft limits"""
        low_limit, high_limit = self.limits

        if low_limit < high_limit and not (low_limit <= pos <= high_limit):
            raise LimitError(f"position={pos} not within limits {self.limits}")



import logging
import time
import pytest
from copy import copy
from unittest import mock
from unittest.mock import Mock

from ophyd.mixins import EpicsSignalPositioner
from ophyd.utils import record_field
from ophyd.status import wait

from .config import motor_recs


logger = logging.getLogger(__name__)


def setUpModule():
    logging.getLogger("ophyd.mixins").setLevel(logging.DEBUG)


def tearDownModule():
    logger.debug("Cleaning up")


@pytest.mark.motorsim
def test_epics_signal_positioner():
    readback = record_field(motor_recs[0], "RBV")
    setpoint = record_field(motor_recs[0], "VAL")
    p = EpicsSignalPositioner(
        readback, write_pv=setpoint, name="p", egu="egu", tolerance=0.005
    )
    p.wait_for_connection()
    assert p.connected

    position_callback = Mock()
    started_motion_callback = Mock()
    finished_motion_callback = Mock()
    moved_cb = Mock()
    assert p.egu == "egu"

    p.subscribe(position_callback, event_type=p.SUB_READBACK)
    p.subscribe(started_motion_callback, event_type=p.SUB_START)
    p.subscribe(finished_motion_callback, event_type=p.SUB_DONE)

    start_pos = p.position
    target_pos = start_pos - 1.5
    p.move(target_pos, wait=True, moved_cb=moved_cb)
    logger.debug(str(p))
    assert not p.moving
    assert abs(p.position - target_pos) <= p.tolerance
    time.sleep(0.5)
    moved_cb.assert_called_with(obj=p)
    position_callback.assert_called_with(
        obj=p, value=mock.ANY, sub_type=p.SUB_READBACK, timestamp=mock.ANY
    )
    started_motion_callback.assert_called_once_with(
        obj=p, sub_type=p.SUB_START, timestamp=mock.ANY
    )
    finished_motion_callback.assert_called_once_with(
        obj=p, sub_type=p.SUB_DONE, value=None, timestamp=mock.ANY
    )

    st = p.set(start_pos)

    wait(st)

    assert st.done
    assert st.error == 0
    assert st.elapsed > 0
    assert abs(p.position - start_pos) <= p.tolerance

    repr(p)
    str(p)

    p.stop()

    p.position

    pc = copy(p)
    assert pc.egu == p.egu
    assert pc.limits == p.limits



import logging
from unittest.mock import Mock

import numpy as np
import pytest

from ophyd import Component, Device, FormattedComponent
from ophyd.device import (
    ComponentWalk,
    create_device_from_components,
    do_not_wait_for_lazy_connection,
    required_for_connection,
    wait_for_lazy_connection,
)
from ophyd.signal import (
    ArrayAttributeSignal,
    AttributeSignal,
    ReadOnlyError,
    Signal,
    SignalRO,
)
from ophyd.utils import ExceptionBundle

logger = logging.getLogger(__name__)


class FakeSignal(Signal):
    def __init__(self, read_pv, *, name=None, parent=None, **kwargs):
        self.read_pv = read_pv
        super().__init__(name=name, parent=parent, **kwargs)
        self._waited_for_connection = False
        self._subscriptions = []

    def wait_for_connection(self):
        self._waited_for_connection = True

    def subscribe(self, method, event_type, **kw):
        self._subscriptions.append((method, event_type, kw))

    def get(self):
        return self.name

    def describe_configuration(self):
        return {self.name + "_conf": {"source": "SIM:test"}}

    def read_configuration(self):
        return {self.name + "_conf": {"value": 0}}


def test_device_state():
    d = Device("test", name="test")

    d.stage()
    old, new = d.configure({})
    assert old == new
    d.unstage()


def test_attrs():
    class MyDevice(Device):
        cpt1 = Component(FakeSignal, "1")
        cpt2 = Component(FakeSignal, "2")
        cpt3 = Component(FakeSignal, "3")

    d = MyDevice(
        "prefix", read_attrs=["cpt1"], configuration_attrs=["cpt2"], name="test"
    )

    d.read()
    assert d.read_attrs == ["cpt1"]
    assert d.configuration_attrs == ["cpt2"]

    assert list(d.read().keys()) == [d.cpt1.name]
    assert set(d.read_configuration().keys()) == {d.cpt2.name + "_conf"}

    assert list(d.describe().keys()) == [d.cpt1.name]
    assert set(d.describe_configuration().keys()) == {
        d.cpt2.name + "_conf",
    }


def test_complexdevice():
    class SubDevice(Device):
        cpt1 = Component(FakeSignal, "1")
        cpt2 = Component(FakeSignal, "2")
        cpt3 = Component(FakeSignal, "3")

    class SubSubDevice(SubDevice):
        cpt4 = Component(FakeSignal, "4")

    class MyDevice(Device):
        sub1 = Component(SubDevice, "1")
        subsub2 = Component(SubSubDevice, "2")
        cpt3 = Component(FakeSignal, "3")

    device = MyDevice("prefix", name="dev")
    device.configuration_attrs = ["sub1", "subsub2.cpt2", "subsub2.cpt4", "cpt3"]
    device.sub1.read_attrs = ["cpt2"]
    device.sub1.configuration_attrs = ["cpt1"]

    assert device.sub1.parent is device
    assert device.subsub2.parent is device
    assert device.cpt3.parent is device

    assert device.sub1.component_names == ("cpt1", "cpt2", "cpt3")
    assert device.subsub2.component_names == ("cpt1", "cpt2", "cpt3", "cpt4")

    with pytest.raises(AttributeError):
        device.component_names.remove("sub1")

    conf_keys = {
        "dev_sub1_cpt1_conf",  # from sub1.*
        # 'dev_sub1_cpt2_conf',   # not in sub1.config_attrs
        "dev_subsub2_cpt2_conf",  # from subsub2.cpt2
        "dev_subsub2_cpt4_conf",  # from subsub2.cpt4
        "dev_cpt3_conf",  # from cpt3
    }

    assert set(device.describe_configuration().keys()) == conf_keys
    assert set(device.read_configuration().keys()) == conf_keys


def test_complexdevice_stop():
    class SubSubDevice(Device):
        cpt4 = Component(FakeSignal, "4")

        def stop(self, *, success=False):
            self.stop_called = True
            self.success = success
            if self.prefix.endswith("_raises_"):
                raise Exception("stop failed for some reason")

    class SubDevice(Device):
        cpt1 = Component(FakeSignal, "1")
        cpt2 = Component(FakeSignal, "2")
        cpt3 = Component(FakeSignal, "3")
        subsub = Component(SubSubDevice, "")

        def stop(self, *, success=False):
            self.stop_called = True
            self.success = success
            super().stop(success=success)

    class MyDevice(Device):
        sub1 = Component(SubDevice, "1")
        sub2 = Component(SubDevice, "_raises_")
        sub3 = Component(SubDevice, "_raises_")
        cpt3 = Component(FakeSignal, "3")

    dev = MyDevice("", name="mydev")
    with pytest.raises(ExceptionBundle) as cm:
        dev.stop()

    ex = cm.value
    assert len(ex.exceptions) == 2
    assert dev.sub1.stop_called
    assert dev.sub2.stop_called
    assert dev.sub3.stop_called
    assert not dev.sub1.success
    assert not dev.sub2.success
    assert not dev.sub3.success

    assert dev.sub1.subsub.stop_called
    assert dev.sub2.subsub.stop_called
    assert dev.sub3.subsub.stop_called
    assert not dev.sub1.subsub.success
    assert not dev.sub2.subsub.success
    assert not dev.sub3.subsub.success

    dev = MyDevice("", name="mydev")
    with pytest.raises(ExceptionBundle) as cm:
        dev.stop(success=True)

    ex = cm.value
    assert len(ex.exceptions) == 2
    assert dev.sub1.stop_called
    assert dev.sub2.stop_called
    assert dev.sub3.stop_called
    assert dev.sub1.success
    assert dev.sub2.success
    assert dev.sub3.success

    assert dev.sub1.subsub.stop_called
    assert dev.sub2.subsub.stop_called
    assert dev.sub3.subsub.stop_called
    assert dev.sub1.subsub.success
    assert dev.sub2.subsub.success
    assert dev.sub3.subsub.success


def test_name_shadowing():
    RESERVED_ATTRS = [
        "name",
        "parent",
        "component_names",
        "_signals",
        "_sig_attrs",
        "_sub_devices",
    ]

    type("a", (Device,), {"a": None})  # legal class definition
    # Illegal class definitions:
    for attr in RESERVED_ATTRS:
        with pytest.raises(TypeError):
            type("a", Device, attr=None)


def test_formatted_component():
    FC = FormattedComponent

    class MyDevice(Device):
        cpt = Component(FakeSignal, "suffix")
        ch = FC(FakeSignal, "{self.prefix}{self._ch}")

        def __init__(self, prefix, ch="a", **kwargs):
            self._ch = ch
            super().__init__(prefix, **kwargs)

    ch_value = "_test_"

    device = MyDevice("prefix:", ch=ch_value, name="test")
    assert device.cpt.parent is device
    assert device.ch.parent is device
    assert device._ch is ch_value
    assert device.ch.read_pv == device.prefix + ch_value
    assert device.cpt.read_pv == device.prefix + MyDevice.cpt.suffix


def test_root():
    class MyDevice(Device):
        cpt = Component(FakeSignal, "suffix")

    d = MyDevice("", name="test")
    assert d.cpt.root == d
    assert d.root == d


def test_hidden_component():
    class MyDevice(Device):
        _hidden_sig = Component(FakeSignal, "suffix")

    d = MyDevice("", name="test")
    assert "_hidden_sig" in d.component_names
    assert not hasattr(d.get(), "_hidden_sig")


def test_attribute_signal():
    init_value = 33

    class SubDevice(Device):
        @property
        def prop(self):
            return init_value

        @prop.setter
        def prop(self, value):
            pass

    class MyDevice(Device):
        sub1 = Component(SubDevice, "1")
        attrsig = Component(AttributeSignal, "prop")
        sub_attrsig = Component(AttributeSignal, "sub1.prop", write_access=False)

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._value = init_value

        @property
        def prop(self):
            return self._value

        @prop.setter
        def prop(self, value):
            self._value = value

    dev = MyDevice("", name="mydev")
    assert dev.describe()["mydev_attrsig"]["source"] == "PY:mydev.prop"
    assert dev.describe()["mydev_sub_attrsig"]["source"] == "PY:mydev.sub1.prop"
    assert dev.attrsig.get() == init_value
    cb = Mock()
    dev.attrsig.subscribe(cb)
    dev.attrsig.put(55)
    assert dev.attrsig.get() == 55
    assert cb.called
    assert dev.sub_attrsig.get() == init_value
    assert dev.sub_attrsig.read_access
    assert not dev.sub_attrsig.write_access

    with pytest.raises(ReadOnlyError):
        dev.sub_attrsig.put(0)


def test_attribute_signal_attributeerror():
    class MyDevice(Device):
        sig1 = Component(AttributeSignal, "unknownattribute")
        sig2 = Component(AttributeSignal, "__class__.__name")
        sig3 = Component(AttributeSignal, "unknown.attribute")

    dev = MyDevice("", name="mydev")
    pytest.raises(AttributeError, dev.sig1.get)
    pytest.raises(AttributeError, dev.sig2.get)
    pytest.raises(AttributeError, dev.sig3.get)
    pytest.raises(AttributeError, getattr, dev.sig3, "base")


def test_shadowing_bs_interface_raises_typeerror():
    class LegalDevice(Device):
        cpt = Component(FakeSignal, "cpt")

    with pytest.raises(TypeError):

        class IllegalDevice(Device):
            # 'read' shadows the bluesky interface and should not be allowed
            read = Component(FakeSignal, "read")


def test_array_attribute_signal():
    init_value = range(10)

    class MyDevice(Device):
        attrsig = Component(ArrayAttributeSignal, "prop")

        @property
        def prop(self):
            return init_value

    dev = MyDevice("", name="mydev")
    np.testing.assert_array_equal(dev.attrsig.get(), init_value)
    assert isinstance(dev.attrsig.get(), np.ndarray)
    assert isinstance(dev.attrsig.get(), np.ndarray)


def test_signal_names():
    class MyDevice(Device):
        cpt = Component(FakeSignal, "suffix")

    d = MyDevice("", name="test")
    with pytest.warns(UserWarning):
        signal_names = d.signal_names

    assert signal_names is d.component_names


def test_summary():
    class MyDevice(Device):
        cpt = Component(FakeSignal, "suffix")

    d = MyDevice("", name="test")
    # smoke tests
    d.summary()
    d.__str__()
    d.__repr__()


def test_labels():
    class MyDevice(Device):
        cpt = Component(FakeSignal, "suffix")

    d = MyDevice("", name="test", labels={"a", "b"})
    assert d._ophyd_labels_ == {"a", "b"}


def test_device_put():
    class MyDevice(Device):
        a = Component(Signal)
        b = Component(Signal)

    d = MyDevice("", name="test")
    d.put((1, 2))
    assert d.get() == (1, 2)

    devtuple = d.get_device_tuple()
    d.put(devtuple(a=10, b=12))
    assert d.get() == (10, 12)

    with pytest.raises(ValueError):
        d.put((1, 2, 3))


def test_lazy_wait_for_connect():
    class MyDevice(Device):
        lazy_wait_for_connection = False
        cpt = Component(FakeSignal, "suffix", lazy=True)

    d = MyDevice("", name="test")
    with wait_for_lazy_connection(d):
        d.cpt

    assert d.cpt._waited_for_connection


def test_lazy_do_not_wait_for_connect():
    class MyDevice(Device):
        lazy_wait_for_connection = True
        cpt = Component(FakeSignal, "suffix", lazy=True)

    d = MyDevice("", name="test")
    with do_not_wait_for_lazy_connection(d):
        d.cpt

    assert not d.cpt._waited_for_connection


def test_sub_decorator():
    class MyDevice(Device):
        cpt = Component(FakeSignal, "suffix", lazy=True)

        @cpt.sub_default
        def default(self, **kw):
            pass

        @cpt.sub_value
        def value(self, **kw):
            pass

        @cpt.sub_meta
        def metadata(self, **kw):
            pass

        @cpt.sub_default
        @cpt.sub_value
        @cpt.sub_meta
        def multi(self, **kw):
            pass

    d = MyDevice("", name="test")

    subs = set(event_type for method, event_type, kw in d.cpt._subscriptions)
    assert subs == {None, "value", "meta"}
    assert len(MyDevice.multi._subscriptions) == 3


def test_walk_components():
    class SubSubDevice(Device):
        cpt4 = Component(FakeSignal, "4")

    class SubDevice(Device):
        cpt1 = Component(FakeSignal, "1")
        cpt2 = Component(FakeSignal, "2")
        cpt3 = Component(FakeSignal, "3")
        subsub = Component(SubSubDevice, "")

    class MyDevice(Device):
        sub1 = Component(SubDevice, "sub1")
        sub2 = Component(SubDevice, "sub2")
        sub3 = Component(SubDevice, "sub3")
        cpt3 = Component(FakeSignal, "cpt3")

    assert list(MyDevice.walk_components()) == [
        ComponentWalk(ancestors=(MyDevice,), dotted_name="sub1", item=MyDevice.sub1),
        ComponentWalk(
            ancestors=(
                MyDevice,
                SubDevice,
            ),
            dotted_name="sub1.cpt1",
            item=MyDevice.sub1.cls.cpt1,
        ),
        ComponentWalk(
            ancestors=(
                MyDevice,
                SubDevice,
            ),
            dotted_name="sub1.cpt2",
            item=MyDevice.sub1.cls.cpt2,
        ),
        ComponentWalk(
            ancestors=(
                MyDevice,
                SubDevice,
            ),
            dotted_name="sub1.cpt3",
            item=MyDevice.sub1.cls.cpt3,
        ),
        ComponentWalk(
            ancestors=(
                MyDevice,
                SubDevice,
            ),
            dotted_name="sub1.subsub",
            item=MyDevice.sub1.cls.subsub,
        ),
        ComponentWalk(
            ancestors=(
                MyDevice,
                SubDevice,
                SubSubDevice,
            ),
            dotted_name="sub1.subsub.cpt4",
            item=MyDevice.sub1.cls.subsub.cls.cpt4,
        ),
        ComponentWalk(ancestors=(MyDevice,), dotted_name="sub2", item=MyDevice.sub2),
        ComponentWalk(
            ancestors=(
                MyDevice,
                SubDevice,
            ),
            dotted_name="sub2.cpt1",
            item=MyDevice.sub2.cls.cpt1,
        ),
        ComponentWalk(
            ancestors=(
                MyDevice,
                SubDevice,
            ),
            dotted_name="sub2.cpt2",
            item=MyDevice.sub2.cls.cpt2,
        ),
        ComponentWalk(
            ancestors=(
                MyDevice,
                SubDevice,
            ),
            dotted_name="sub2.cpt3",
            item=MyDevice.sub2.cls.cpt3,
        ),
        ComponentWalk(
            ancestors=(
                MyDevice,
                SubDevice,
            ),
            dotted_name="sub2.subsub",
            item=MyDevice.sub2.cls.subsub,
        ),
        ComponentWalk(
            ancestors=(
                MyDevice,
                SubDevice,
                SubSubDevice,
            ),
            dotted_name="sub2.subsub.cpt4",
            item=MyDevice.sub2.cls.subsub.cls.cpt4,
        ),
        ComponentWalk(ancestors=(MyDevice,), dotted_name="sub3", item=MyDevice.sub3),
        ComponentWalk(
            ancestors=(
                MyDevice,
                SubDevice,
            ),
            dotted_name="sub3.cpt1",
            item=MyDevice.sub3.cls.cpt1,
        ),
        ComponentWalk(
            ancestors=(
                MyDevice,
                SubDevice,
            ),
            dotted_name="sub3.cpt2",
            item=MyDevice.sub3.cls.cpt2,
        ),
        ComponentWalk(
            ancestors=(
                MyDevice,
                SubDevice,
            ),
            dotted_name="sub3.cpt3",
            item=MyDevice.sub3.cls.cpt3,
        ),
        ComponentWalk(
            ancestors=(
                MyDevice,
                SubDevice,
            ),
            dotted_name="sub3.subsub",
            item=MyDevice.sub3.cls.subsub,
        ),
        ComponentWalk(
            ancestors=(
                MyDevice,
                SubDevice,
                SubSubDevice,
            ),
            dotted_name="sub3.subsub.cpt4",
            item=MyDevice.sub3.cls.subsub.cls.cpt4,
        ),
        ComponentWalk(ancestors=(MyDevice,), dotted_name="cpt3", item=MyDevice.cpt3),
    ]


@pytest.mark.parametrize("include_lazy", [False, True])
def test_walk_signals(include_lazy):
    class SubSubDevice(Device):
        cpt4 = Component(FakeSignal, "4", lazy=True)

    class SubDevice(Device):
        cpt1 = Component(FakeSignal, "1")
        cpt2 = Component(FakeSignal, "2")
        cpt3 = Component(FakeSignal, "3")
        subsub = Component(SubSubDevice, "")

    class MyDevice(Device):
        sub1 = Component(SubDevice, "sub1")
        sub2 = Component(SubDevice, "sub2")
        sub3 = Component(SubDevice, "sub3")
        cpt3 = Component(FakeSignal, "cpt3")

    print(MyDevice.sub1.cls.cpt1)

    dev = MyDevice("", name="mydev")
    walked_list = list(dev.walk_signals(include_lazy=include_lazy))

    dev.summary()
    expected = [
        ComponentWalk(
            ancestors=(
                dev,
                dev.sub1,
            ),
            dotted_name="sub1.cpt1",
            item=dev.sub1.cpt1,
        ),
        ComponentWalk(
            ancestors=(
                dev,
                dev.sub1,
            ),
            dotted_name="sub1.cpt2",
            item=dev.sub1.cpt2,
        ),
        ComponentWalk(
            ancestors=(
                dev,
                dev.sub1,
            ),
            dotted_name="sub1.cpt3",
            item=dev.sub1.cpt3,
        ),
        ComponentWalk(
            ancestors=(
                dev,
                dev.sub1,
                dev.sub1.subsub,
            ),
            dotted_name="sub1.subsub.cpt4",
            item=dev.sub1.subsub.cpt4,
        ),
        ComponentWalk(
            ancestors=(
                dev,
                dev.sub2,
            ),
            dotted_name="sub2.cpt1",
            item=dev.sub2.cpt1,
        ),
        ComponentWalk(
            ancestors=(
                dev,
                dev.sub2,
            ),
            dotted_name="sub2.cpt2",
            item=dev.sub2.cpt2,
        ),
        ComponentWalk(
            ancestors=(
                dev,
                dev.sub2,
            ),
            dotted_name="sub2.cpt3",
            item=dev.sub2.cpt3,
        ),
        ComponentWalk(
            ancestors=(
                dev,
                dev.sub2,
                dev.sub2.subsub,
            ),
            dotted_name="sub2.subsub.cpt4",
            item=dev.sub2.subsub.cpt4,
        ),
        ComponentWalk(
            ancestors=(
                dev,
                dev.sub3,
            ),
            dotted_name="sub3.cpt1",
            item=dev.sub3.cpt1,
        ),
        ComponentWalk(
            ancestors=(
                dev,
                dev.sub3,
            ),
            dotted_name="sub3.cpt2",
            item=dev.sub3.cpt2,
        ),
        ComponentWalk(
            ancestors=(
                dev,
                dev.sub3,
            ),
            dotted_name="sub3.cpt3",
            item=dev.sub3.cpt3,
        ),
        ComponentWalk(
            ancestors=(
                dev,
                dev.sub3,
                dev.sub3.subsub,
            ),
            dotted_name="sub3.subsub.cpt4",
            item=dev.sub3.subsub.cpt4,
        ),
        ComponentWalk(ancestors=(dev,), dotted_name="cpt3", item=dev.cpt3),
    ]

    if not include_lazy:
        expected = [item for item in expected if "cpt4" not in item.dotted_name]

    assert walked_list == expected


def test_walk_subdevice_classes():
    class SubSubDevice(Device):
        cpt4 = Component(FakeSignal, "4")

    class SubDevice(Device):
        cpt1 = Component(FakeSignal, "1")
        cpt2 = Component(FakeSignal, "2")
        cpt3 = Component(FakeSignal, "3")
        subsub = Component(SubSubDevice, "")

    class MyDevice(Device):
        sub1 = Component(SubDevice, "sub1")
        sub2 = Component(SubDevice, "sub2")
        sub3 = Component(SubDevice, "sub3")
        cpt3 = Component(FakeSignal, "cpt3")

    assert list(MyDevice.walk_subdevice_classes()) == [
        ("sub1", SubDevice),
        ("sub1.subsub", SubSubDevice),
        ("sub2", SubDevice),
        ("sub2.subsub", SubSubDevice),
        ("sub3", SubDevice),
        ("sub3.subsub", SubSubDevice),
    ]


def test_walk_subdevices():
    class SubSubDevice(Device):
        cpt4 = Component(FakeSignal, "4")

    class SubDevice(Device):
        cpt1 = Component(FakeSignal, "1")
        cpt2 = Component(FakeSignal, "2")
        cpt3 = Component(FakeSignal, "3")
        subsub = Component(SubSubDevice, "")

    class MyDevice(Device):
        sub1 = Component(SubDevice, "sub1")
        sub2 = Component(SubDevice, "sub2")
        sub3 = Component(SubDevice, "sub3")
        cpt3 = Component(FakeSignal, "cpt3")

    dev = MyDevice("", name="mydev")
    assert list(dev.walk_subdevices()) == [
        ("sub1", dev.sub1),
        ("sub1.subsub", dev.sub1.subsub),
        ("sub2", dev.sub2),
        ("sub2.subsub", dev.sub2.subsub),
        ("sub3", dev.sub3),
        ("sub3.subsub", dev.sub3.subsub),
    ]


def test_dotted_name():
    from ophyd import Component as Cpt
    from ophyd import Device
    from ophyd.sim import SynSignal

    class Inner(Device):
        x = Cpt(SynSignal)
        y = Cpt(SynSignal)

    class Outer(Device):
        a = Cpt(Inner)
        b = Cpt(Inner)

    o = Outer(name="test")

    assert o.dotted_name == ""
    assert o.a.dotted_name == "a"
    assert o.b.dotted_name == "b"

    assert o.a.x.dotted_name == "a.x"
    assert o.b.x.dotted_name == "b.x"

    assert o.a.y.dotted_name == "a.y"
    assert o.b.y.dotted_name == "b.y"

    assert o.attr_name == ""
    assert o.a.attr_name == "a"
    assert o.b.attr_name == "b"

    assert o.a.x.attr_name == "x"
    assert o.b.x.attr_name == "x"

    assert o.a.y.attr_name == "y"
    assert o.b.y.attr_name == "y"


def test_create_device():
    components = dict(
        cpt1=Component(Signal, value=0),
        cpt2=Component(Signal, value=1),
        cpt3=Component(Signal, value=2),
    )
    Dev = create_device_from_components("Dev", base_class=Device, **components)
    assert Dev.__name__ == "Dev"
    assert Dev.cpt1 is components["cpt1"]
    dev = Dev(name="dev")
    assert dev.cpt1.get() == 0
    assert dev.cpt2.get() == 1
    assert dev.cpt3.get() == 2


def test_create_device_bad_component():
    with pytest.raises(ValueError):
        create_device_from_components("Dev", base_class=Device, bad_component=None)


def test_required_for_connection_on_method_with_subscriptions():
    class MyDevice(Device):
        cpt = Component(Signal, value=0)

        @required_for_connection
        @cpt.sub_value
        def method(self):
            ...

    dev = MyDevice(name="dev")

    with pytest.raises(TimeoutError):
        dev.wait_for_connection(timeout=0.1)

    dev.cpt.put(0)
    dev.wait_for_connection(timeout=0.1)


def test_required_for_connection_on_method():
    class MyDevice(Device):
        @required_for_connection
        def method(self):
            ...

    dev = MyDevice(name="dev")

    # Timeout without it having been called:
    with pytest.raises(TimeoutError):
        dev.wait_for_connection(timeout=0.01)

    # Call and expect no timeout:
    dev.method()
    dev.wait_for_connection(timeout=0.01)


def test_required_for_connection_in_init():
    class MyDevice(Device):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.call_to_connect = required_for_connection(self.method, device=self)

        def method(self):
            print("method called")

    dev = MyDevice(name="dev")

    # Timeout without it having been called:
    with pytest.raises(TimeoutError):
        dev.wait_for_connection(timeout=0.01)

    # Call and expect no timeout:
    dev.call_to_connect()
    dev.wait_for_connection(timeout=0.01)


def test_noneified_component():
    class SubDevice(Device):
        cpt1 = Component(FakeSignal, "1")

    class MyDevice(Device):
        sub1 = Component(SubDevice, "sub1")
        sub2 = Component(SubDevice, "sub2")

    class MyDeviceWithoutSub2(Device):
        sub1 = Component(SubDevice, "sub1")
        sub2 = None

    assert MyDevice.component_names == ("sub1", "sub2")
    assert MyDevice._sub_devices == ["sub1", "sub2"]

    assert MyDeviceWithoutSub2.component_names == ("sub1",)
    assert MyDeviceWithoutSub2._sub_devices == ["sub1"]


@pytest.mark.parametrize(
    "lazy_state, cntx",
    [(False, wait_for_lazy_connection), (True, do_not_wait_for_lazy_connection)],
)
def test_lazy_wait_context(lazy_state, cntx):
    class LocalExcepton(Exception):
        ...

    d = Device(name="d")
    d.lazy_wait_for_connection = lazy_state

    try:
        with cntx(d):
            assert d.lazy_wait_for_connection == (not lazy_state)
            raise LocalExcepton
    except LocalExcepton:
        ...

    assert d.lazy_wait_for_connection is lazy_state


def test_non_Divice_mixin_with_components():
    class Host(Device):
        a = Component(Signal)

    class MixIn:
        b = Component(Signal)

    class Target(Host, MixIn):
        ...

    t = Target(name="target")

    with pytest.raises(RuntimeError):
        t.b


def test_annotated_device():
    class MyDevice(Device):
        cpt1 = Component[Signal](Signal)
        cpt2 = Component[SignalRO](SignalRO)
        cpt3 = Component[SignalRO](SignalRO)
        cpt4 = Component(SignalRO)

    dev = MyDevice(name="dev")
    assert isinstance(dev.cpt1, Signal)
    assert isinstance(dev.cpt2, SignalRO)
    assert MyDevice.cpt1._get_class_from_annotation() is Signal
    assert MyDevice.cpt2._get_class_from_annotation() is SignalRO
    assert MyDevice.cpt3._get_class_from_annotation() is SignalRO
    assert MyDevice.cpt3.cls is SignalRO
    assert MyDevice.cpt4._get_class_from_annotation() is None



import copy
import inspect
import itertools
from functools import partial

import numpy as np
import os
import random
import threading
import time as ttime
import uuid
import weakref
import warnings

from collections import deque, OrderedDict
from tempfile import mkdtemp

from .signal import Signal, EpicsSignal, EpicsSignalRO
from .areadetector.base import EpicsSignalWithRBV
from .areadetector.paths import EpicsPathSignal
from .status import DeviceStatus, StatusBase
from .device import Device, Component as Cpt, DynamicDeviceComponent as DDCpt, Kind
from types import SimpleNamespace
from .pseudopos import (
    PseudoPositioner,
    PseudoSingle,
    real_position_argument,
    pseudo_position_argument,
)
from .positioner import SoftPositioner
from .utils import ReadOnlyError, LimitError
from .log import logger


# two convenience functions 'vendored' from bluesky.utils


def new_uid():
    return str(uuid.uuid4())


def short_uid(label=None, truncate=6):
    "Return a readable but unique id like 'label-fjfi5a'"
    if label:
        return "-".join([label, new_uid()[:truncate]])
    else:
        return new_uid()[:truncate]


class NullStatus(StatusBase):
    "A simple Status object that is always immediately done, successfully."

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_finished()
        self.wait()


class EnumSignal(Signal):
    def __init__(self, *args, value=0, enum_strings, **kwargs):
        super().__init__(*args, value=0, **kwargs)
        self._enum_strs = tuple(enum_strings)
        self._metadata["enum_strs"] = tuple(enum_strings)
        self.put(value)

    def put(self, value, **kwargs):
        if value in self._enum_strs:
            value = self._enum_strs.index(value)
        elif isinstance(value, str):
            err = f"{value} not in enum strs {self._enum_strs}"
            raise ValueError(err)
        return super().put(value, **kwargs)

    def get(self, *, as_string=True, **kwargs):
        """
        Implement getting as enum strings
        """
        value = super().get()

        if as_string:
            if self._enum_strs is not None and isinstance(value, int):
                return self._enum_strs[value]
            elif value is not None:
                return str(value)
        return value

    def describe(self):
        desc = super().describe()
        desc[self.name]["enum_strs"] = self._enum_strs
        return desc


class SynSignal(Signal):
    """
    A synthetic Signal that evaluates a Python function when triggered.

    Parameters
    ----------
    func : callable, optional
        This function sets the signal to a new value when it is triggered.
        Expected signature: ``f() -> value``.
        By default, triggering the signal does not change the value.
    name : string, keyword only
    exposure_time : number, optional
        Seconds of delay when triggered (simulated 'exposure time'). Default is
        0.
    precision : integer, optional
        Digits of precision. Default is 3.
    parent : Device, optional
        Used internally if this Signal is made part of a larger Device.
    kind : a member the Kind IntEnum (or equivalent integer), optional
        Default is Kind.normal. See Kind for options.

    """

    # This signature is arranged to mimic the signature of EpicsSignal, where
    # the Python function (func) takes the place of the PV.
    def __init__(
        self,
        func=None,
        *,
        name,  # required, keyword-only
        exposure_time=0,
        precision=3,
        parent=None,
        labels=None,
        kind=None,
        **kwargs,
    ):
        if func is None:
            # When triggered, just put the current value.
            func = self.get
            # Initialize readback with 0.
            self._readback = 0
        sentinel = object()
        loop = kwargs.pop("loop", sentinel)
        if loop is not sentinel:
            warnings.warn(
                f"{self.__class__} no longer takes a loop as input.  "
                "Your input will be ignored and may raise in the future",
                stacklevel=2,
            )
        self._func = func
        self.exposure_time = exposure_time
        self.precision = precision
        super().__init__(
            value=self._func(),
            timestamp=ttime.time(),
            name=name,
            parent=parent,
            labels=labels,
            kind=kind,
            **kwargs,
        )
        self._metadata.update(
            connected=True,
        )

    def describe(self):
        res = super().describe()
        # There should be only one key here, but for the sake of generality....
        for k in res:
            res[k]["precision"] = self.precision
        return res

    def trigger(self):
        st = DeviceStatus(device=self)
        delay_time = self.exposure_time
        if delay_time:

            def sleep_and_finish():
                self.log.debug("sleep_and_finish %s", self)
                ttime.sleep(delay_time)
                self.put(self._func())
                st.set_finished()

            threading.Thread(target=sleep_and_finish, daemon=True).start()
        else:
            self.put(self._func())
            st.set_finished()
        return st

    def sim_set_func(self, func):
        """
        Update the SynSignal function to set a new value on trigger.
        """
        self._func = func


class SynSignalRO(SynSignal):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._metadata.update(
            connected=True,
            write_access=False,
        )

    def put(self, value, *, timestamp=None, force=False):
        msg = f"{self}.put(value={value}, timestamp={timestamp}, force={force})"
        self.log.error(msg)
        raise ReadOnlyError(msg)

    def set(self, value, *, timestamp=None, force=False):
        msg = f"{self} is readonly"
        self.log.error(msg)
        raise ReadOnlyError(msg)


class SynPeriodicSignal(SynSignal):
    """
    A synthetic Signal that evaluates a Python function periodically.
    The signal value is updated in a background thread. To start the thread,
    call the `start_simulation()` method before the beginning of simulation.

    Parameters
    ----------
    func : callable, optional
        This function sets the signal to a new value when it is triggered.
        Expected signature: ``f() -> value``.
        By default, triggering the signal generates white noise on [0, 1].
    name : string, keyword only
    period : number, optional
        How often the Signal's value is updated in the background. Default is
        1 second.
    period_jitter : number, optional
        Random Gaussian variation of the period. Default is 1 second.
    exposure_time : number, optional
        Seconds of delay when triggered (simulated 'exposure time'). Default is
        0.
    parent : Device, optional
        Used internally if this Signal is made part of a larger Device.
    kind : a member the Kind IntEnum (or equivalent integer), optional
        Default is Kind.normal. See Kind for options.
    """

    def __init__(
        self,
        func=None,
        *,
        name,  # required, keyword-only
        period=1,
        period_jitter=1,
        exposure_time=0,
        parent=None,
        labels=None,
        kind=None,
        **kwargs,
    ):
        if func is None:
            func = np.random.rand

        self._period = period
        self._period_jitter = period_jitter

        super().__init__(
            name=name,
            func=func,
            exposure_time=exposure_time,
            parent=parent,
            labels=labels,
            kind=kind,
            **kwargs,
        )
        self.__thread = None

    def start_simulation(self):
        """
        Start background thread that performs periodic value updates. The method
        should be called at least once before the beginning of simulation. Multiple
        calls to the method are ignored.
        """
        if self.__thread is None:

            def periodic_update(ref, period, period_jitter):
                while True:
                    signal = ref()
                    if not signal:
                        # Our target Signal has been garbage collected. Shut
                        # down the Thread.
                        return
                    signal.put(signal._func())
                    del signal
                    # Sleep for period +/- period_jitter.
                    ttime.sleep(
                        max(self._period + self._period_jitter * np.random.randn(), 0)
                    )

            self.__thread = threading.Thread(
                target=periodic_update,
                daemon=True,
                args=(weakref.ref(self), self._period, self._period_jitter),
            )
            self.__thread.start()

    def _start_simulation_deprecated(self):
        """Call `start_simulation` and print deprecation warning."""
        if self.__thread is None:
            msg = (
                "Deprecated API: Objects of SynPeriodicSignal must be initialized before simulation\n"
                "by calling 'start_simulation()' method. Two such objects ('rand' and 'rand2') are\n"
                "created by 'ophyd.sim' module. Call\n"
                "    rand.start_simulation() or rand2.start_simulation()\n"
                "before the object is used."
            )
            self.log.warning(msg)
            self.start_simulation()

    def trigger(self):
        self._start_simulation_deprecated()
        return super().trigger()

    def get(self, **kwargs):
        self._start_simulation_deprecated()
        return super().get(**kwargs)

    def put(self, *args, **kwargs):
        self._start_simulation_deprecated()
        super().put(*args, **kwargs)

    def set(self, *args, **kwargs):
        self._start_simulation_deprecated()
        return super().set(*args, **kwargs)

    def read(self):
        self._start_simulation_deprecated()
        return super().read()

    def subscribe(self, *args, **kwargs):
        self._start_simulation_deprecated()
        return super().subscribe(*args, **kwargs)


class _ReadbackSignal(Signal):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._metadata.update(
            connected=True,
            write_access=False,
        )

    def get(self):
        self._readback = self.parent.sim_state["readback"]
        return self._readback

    def describe(self):
        res = super().describe()
        # There should be only one key here, but for the sake of
        # generality....
        for k in res:
            res[k]["precision"] = self.parent.precision
        return res

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        return self.parent.sim_state["readback_ts"]

    def put(self, value, *, timestamp=None, force=False):
        raise ReadOnlyError("The signal {} is readonly.".format(self.name))

    def set(self, value, *, timestamp=None, force=False):
        raise ReadOnlyError("The signal {} is readonly.".format(self.name))


class _SetpointSignal(Signal):
    def put(self, value, *, timestamp=None, force=False):
        self._readback = float(value)
        self.parent.set(float(value))

    def get(self):
        self._readback = self.parent.sim_state["setpoint"]
        return self.parent.sim_state["setpoint"]

    def describe(self):
        res = super().describe()
        # There should be only one key here, but for the sake of generality....
        for k in res:
            res[k]["precision"] = self.parent.precision
        return res

    @property
    def timestamp(self):
        """Timestamp of the readback value"""
        return self.parent.sim_state["setpoint_ts"]


class SynAxis(Device):
    """
    A synthetic settable Device mimic any 1D Axis (position, temperature).

    Parameters
    ----------
    name : string, keyword only
    readback_func : callable, optional
        When the Device is set to ``x``, its readback will be updated to
        ``f(x)``. This can be used to introduce random noise or a systematic
        offset.
        Expected signature: ``f(x) -> value``.
    value : object, optional
        The initial value. Default is 0.
    delay : number, optional
        Simulates how long it takes the device to "move". Default is 0 seconds.
    precision : integer, optional
        Digits of precision. Default is 3.
    parent : Device, optional
        Used internally if this Signal is made part of a larger Device.
    kind : a member the Kind IntEnum (or equivalent integer), optional
        Default is Kind.normal. See Kind for options.
    """

    readback = Cpt(_ReadbackSignal, value=0, kind="hinted")
    setpoint = Cpt(_SetpointSignal, value=0, kind="normal")

    velocity = Cpt(Signal, value=1, kind="config")
    acceleration = Cpt(Signal, value=1, kind="config")

    unused = Cpt(Signal, value=1, kind="omitted")

    SUB_READBACK = "readback"
    _default_sub = SUB_READBACK

    def __init__(
        self,
        *,
        name,
        readback_func=None,
        value=0,
        delay=0,
        precision=3,
        parent=None,
        labels=None,
        kind=None,
        **kwargs,
    ):
        if readback_func is None:

            def readback_func(x):
                return x

        sentinel = object()
        loop = kwargs.pop("loop", sentinel)
        if loop is not sentinel:
            warnings.warn(
                f"{self.__class__} no longer takes a loop as input.  "
                "Your input will be ignored and may raise in the future",
                stacklevel=2,
            )
        self.sim_state = {}
        self._readback_func = readback_func
        self.delay = delay
        self.precision = precision

        # initialize values
        self.sim_state["setpoint"] = value
        self.sim_state["setpoint_ts"] = ttime.time()
        self.sim_state["readback"] = readback_func(value)
        self.sim_state["readback_ts"] = ttime.time()

        super().__init__(name=name, parent=parent, labels=labels, kind=kind, **kwargs)
        self.readback.name = self.name

    def set(self, value):
        old_setpoint = self.sim_state["setpoint"]
        self.sim_state["setpoint"] = value
        self.sim_state["setpoint_ts"] = ttime.time()
        self.setpoint._run_subs(
            sub_type=self.setpoint.SUB_VALUE,
            old_value=old_setpoint,
            value=self.sim_state["setpoint"],
            timestamp=self.sim_state["setpoint_ts"],
        )

        def update_state():
            old_readback = self.sim_state["readback"]
            self.sim_state["readback"] = self._readback_func(value)
            self.sim_state["readback_ts"] = ttime.time()
            self.readback._run_subs(
                sub_type=self.readback.SUB_VALUE,
                old_value=old_readback,
                value=self.sim_state["readback"],
                timestamp=self.sim_state["readback_ts"],
            )
            self._run_subs(
                sub_type=self.SUB_READBACK,
                old_value=old_readback,
                value=self.sim_state["readback"],
                timestamp=self.sim_state["readback_ts"],
            )

        st = DeviceStatus(device=self)
        if self.delay:

            def sleep_and_finish():
                ttime.sleep(self.delay)
                update_state()
                st.set_finished()

            threading.Thread(target=sleep_and_finish, daemon=True).start()
        else:
            update_state()
            st.set_finished()
        return st

    @property
    def position(self):
        return self.readback.get()


class SynAxisEmptyHints(SynAxis):
    @property
    def hints(self):
        return {}


class SynAxisNoHints(SynAxis):
    readback = Cpt(_ReadbackSignal, value=0, kind="omitted")

    @property
    def hints(self):
        raise AttributeError


class SynGauss(Device):
    """
    Evaluate a point on a Gaussian based on the value of a motor.

    Parameters
    ----------
    name : string
    motor : Device
    motor_field : string
    center : number
        center of peak
    Imax : number
        max intensity of peak
    sigma : number, optional
        Default is 1.
    noise : {'poisson', 'uniform', None}, optional
        Add noise to the gaussian peak.
    noise_multiplier : float, optional
        Only relevant for 'uniform' noise. Multiply the random amount of
        noise by 'noise_multiplier'
    random_state : numpy random state object, optional
        np.random.RandomState(0), to generate random number with given seed

    Example
    -------
    motor = SynAxis(name='motor')
    det = SynGauss('det', motor, 'motor', center=0, Imax=1, sigma=1)
    """

    def _compute(self):
        m = self._motor.read()[self._motor_field]["value"]
        # we need to do this one at a time because
        #   - self.read() may be screwed with by the user
        #   - self.get() would cause infinite recursion
        Imax = self.Imax.get()
        center = self.center.get()
        sigma = self.sigma.get()
        noise = self.noise.get()
        noise_multiplier = self.noise_multiplier.get()
        v = Imax * np.exp(-((m - center) ** 2) / (2 * sigma**2))
        if noise == "poisson":
            v = int(self.random_state.poisson(np.round(v), 1))
        elif noise == "uniform":
            v += self.random_state.uniform(-1, 1) * noise_multiplier
        return v

    val = Cpt(SynSignal, kind="hinted")
    Imax = Cpt(Signal, value=10, kind="config")
    center = Cpt(Signal, value=0, kind="config")
    sigma = Cpt(Signal, value=1, kind="config")
    noise = Cpt(
        EnumSignal,
        value="none",
        kind="config",
        enum_strings=("none", "poisson", "uniform"),
    )
    noise_multiplier = Cpt(Signal, value=1, kind="config")

    def __init__(
        self, name, motor, motor_field, center, Imax, *, random_state=None, **kwargs
    ):
        set_later = {}
        for k in ("sigma", "noise", "noise_multiplier"):
            v = kwargs.pop(k, None)
            if v is not None:
                set_later[k] = v
        super().__init__(name=name, **kwargs)
        self._motor = motor
        self._motor_field = motor_field
        self.center.put(center)
        self.Imax.put(Imax)

        self.random_state = random_state or np.random
        self.val.name = self.name
        self.val.sim_set_func(self._compute)
        for k, v in set_later.items():
            getattr(self, k).put(v)

        self.trigger()

    def subscribe(self, *args, **kwargs):
        return self.val.subscribe(*args, **kwargs)

    def clear_sub(self, cb, event_type=None):
        return self.val.clear_sub(cb, event_type=event_type)

    def unsubscribe(self, cid):
        return self.val.unsubscribe(cid)

    def unsubscribe_all(self):
        return self.val.unsubscribe_all()

    def trigger(self, *args, **kwargs):
        return self.val.trigger(*args, **kwargs)

    @property
    def precision(self):
        return self.val.precision

    @precision.setter
    def precision(self, v):
        self.val.precision = v

    @property
    def exposure_time(self):
        return self.val.exposure_time

    @exposure_time.setter
    def exposure_time(self, v):
        self.val.exposure_time = v


class Syn2DGauss(Device):
    """
    Evaluate a point on a Gaussian based on the value of a motor.

    Parameters
    ----------
    name : str
        The name of the detector
    motor0 : SynAxis
        The 'x' coordinate of the 2-D gaussian blob
    motor_field0 : str
        The name field of the motor. Should be the key in motor0.describe()
    motor1 : SynAxis
        The 'y' coordinate of the 2-D gaussian blob
    motor_field1 : str
        The name field of the motor. Should be the key in motor1.describe()
    center : iterable, optional
        The center of the gaussian blob
        Defaults to (0,0)
    Imax : float, optional
        The intensity at `center`
        Defaults to 1
    sigma : float, optional
        Standard deviation for gaussian blob
        Defaults to 1
    noise : {'poisson', 'uniform', None}, optional
        Add noise to the gaussian peak..
        Defaults to None
    noise_multiplier : float, optional
        Only relevant for 'uniform' noise. Multiply the random amount of
        noise by 'noise_multiplier'
        Defaults to 1
    random_state : numpy random state object, optional
        np.random.RandomState(0), to generate random number with given seed

    Example
    -------
    motor = SynAxis(name='motor')
    det = SynGauss('det', motor, 'motor', center=0, Imax=1, sigma=1)
    """

    val = Cpt(SynSignal, kind="hinted")
    Imax = Cpt(Signal, value=10, kind="config")
    center = Cpt(Signal, value=0, kind="config")
    sigma = Cpt(Signal, value=1, kind="config")
    noise = Cpt(
        EnumSignal,
        value="none",
        kind="config",
        enum_strings=("none", "poisson", "uniform"),
    )
    noise_multiplier = Cpt(Signal, value=1, kind="config")

    def _compute(self):
        x = self._motor0.read()[self._motor_field0]["value"]
        y = self._motor1.read()[self._motor_field1]["value"]
        m = np.array([x, y])
        Imax = self.Imax.get()
        center = self.center.get()
        sigma = self.sigma.get()
        noise = self.noise.get()
        noise_multiplier = self.noise_multiplier.get()
        v = Imax * np.exp(-np.sum((m - center) ** 2) / (2 * sigma**2))
        if noise == "poisson":
            v = int(self.random_state.poisson(np.round(v), 1))
        elif noise == "uniform":
            v += self.random_state.uniform(-1, 1) * noise_multiplier
        return v

    def __init__(
        self,
        name,
        motor0,
        motor_field0,
        motor1,
        motor_field1,
        center,
        Imax,
        sigma=1,
        noise="none",
        noise_multiplier=1,
        random_state=None,
        **kwargs,
    ):
        super().__init__(name=name, **kwargs)
        self._motor0 = motor0
        self._motor1 = motor1
        self._motor_field0 = motor_field0
        self._motor_field1 = motor_field1
        self.center.put(center)
        self.Imax.put(Imax)
        self.sigma.put(sigma)
        self.noise.put(noise)
        self.noise_multiplier.put(noise_multiplier)

        if random_state is None:
            random_state = np.random
        self.random_state = random_state
        self.val.name = self.name
        self.val.sim_set_func(self._compute)

        self.trigger()

    def trigger(self, *args, **kwargs):
        return self.val.trigger(*args, **kwargs)


class TrivialFlyer:
    """Trivial flyer that complies to the API but returns empty data."""

    name = "trivial_flyer"
    parent = None

    def kickoff(self):
        return NullStatus()

    def describe_collect(self):
        return {"stream_name": {}}

    def read_configuration(self):
        return OrderedDict()

    def describe_configuration(self):
        return OrderedDict()

    def complete(self):
        return NullStatus()

    def collect(self):
        for i in range(100):
            yield {"data": {}, "timestamps": {}, "time": i, "seq_num": i}

    def stop(self, *, success=False):
        pass


class NewTrivialFlyer(TrivialFlyer):
    """
    The old-style API inserted Resource and Datum documents into a database
    directly. The new-style API only caches the documents and provides an
    interface (collect_asset_docs) for accessing that cache. This change was
    part of the "asset refactor" that changed that way Resource and Datum
    documents flowed through ophyd, bluesky, and databroker. Trivial flyer that
    complies to the API but returns empty data.
    """

    name = "new_trivial_flyer"

    def collect_asset_docs(self):
        for _ in ():
            yield _


class MockFlyer:
    """
    Class for mocking a flyscan API implemented with stepper motors.
    """

    def __init__(self, name, detector, motor, start, stop, num, **kwargs):
        self.name = name
        self.parent = None
        self._mot = motor
        self._detector = detector
        self._steps = np.linspace(start, stop, num)
        self._data = deque()
        self._completion_status = None
        self._lock = threading.RLock()
        sentinel = object()
        loop = kwargs.pop("loop", sentinel)
        if loop is not sentinel:
            warnings.warn(
                f"{self.__class__} no longer takes a loop as input.  "
                "Your input will be ignored and may raise in the future",
                stacklevel=2,
            )
        if kwargs:
            raise TypeError(
                f"{self.__class__}.__init__ got unexpected "
                f"keyword arguments {list(kwargs)}"
            )

    def __setstate__(self, val):
        name, detector, motor, steps = val
        self.name = name
        self.parent = None
        self._mot = motor
        self._detector = detector
        self._steps = steps
        self._completion_status = None

    def __getstate__(self):
        return (self.name, self._detector, self._mot, self._steps)

    def read_configuration(self):
        return {}

    def describe_configuration(self):
        return {}

    def describe_collect(self):
        dd = dict()
        dd.update(self._mot.describe())
        dd.update(self._detector.describe())
        return {self.name: dd}

    def complete(self):
        if self._completion_status is None:
            raise RuntimeError("No collection in progress")
        return self._completion_status

    def kickoff(self):
        if self._completion_status is not None and not self._completion_status.done:
            raise RuntimeError("Kicking off a second time?!")
        self._data = deque()
        st = DeviceStatus(device=self)
        self._completion_status = st

        def flyer_worker():
            self._scan()
            st.set_finished()

        threading.Thread(target=flyer_worker, daemon=True).start()
        kickoff_st = DeviceStatus(device=self)
        kickoff_st.set_finished()
        return kickoff_st

    def collect(self):

        with self._lock:
            data = list(self._data)
            self._data.clear()
        yield from data

    def _scan(self):
        "This will be run on a separate thread, started in self.kickoff()"
        ttime.sleep(0.1)
        for p in self._steps:
            stat = self._mot.set(p)
            stat.wait()
            stat = self._detector.trigger()
            stat.wait()

            event = dict()
            event["time"] = ttime.time()
            event["data"] = dict()
            event["timestamps"] = dict()
            for r in [self._mot, self._detector]:
                d = r.read()
                for k, v in d.items():
                    event["data"][k] = v["value"]
                    event["timestamps"][k] = v["timestamp"]
            with self._lock:
                self._data.append(event)

    def stop(self, *, success=False):
        pass


class SynSignalWithRegistry(SynSignal):
    """
    A SynSignal integrated with databroker.assets

    Parameters
    ----------
    func : callable, optional
        This function sets the signal to a new value when it is triggered.
        Expected signature: ``f() -> value``.
        By default, triggering the signal does not change the value.
    name : string, keyword only
    exposure_time : number, optional
        Seconds of delay when triggered (simulated 'exposure time'). Default is
        0.
    parent : Device, optional
        Used internally if this Signal is made part of a larger Device.
    reg : Registry, optional
        DEPRECATED. If used, this is ignored and a warning is issued. In a
        future release, this parameter will be removed.
    save_path : str, optional
        Path to save files to, if None make a temp dir, defaults to None.
    save_func : function, optional
        The function to save the data, function signature must be:
        `func(file_path, array)`, defaults to np.save
    save_spec : str, optional
        The spec for the save function, defaults to 'RWFS_NPY'
    save_ext : str, optional
        The extension to add to the file name, defaults to '.npy'

    """

    def __init__(
        self,
        *args,
        save_path=None,
        save_func=partial(np.save, allow_pickle=False),
        save_spec="NPY_SEQ",
        save_ext="npy",
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.save_func = save_func
        self.save_ext = save_ext
        self._resource_uid = None
        self._datum_counter = None
        self._asset_docs_cache = deque()
        if save_path is None:
            self.save_path = mkdtemp()
        else:
            self.save_path = save_path
        self._spec = save_spec  # spec name stored in resource doc

        self._file_stem = None
        self._path_stem = None
        self._result = {}

    def stage(self):
        self._file_stem = short_uid()
        self._datum_counter = itertools.count()
        self._path_stem = os.path.join(self.save_path, self._file_stem)

        # This is temporarily more complicated than it will be in the future.
        # It needs to support old configurations that have a registry.
        resource = {
            "spec": self._spec,
            "root": self.save_path,
            "resource_path": self._file_stem,
            "resource_kwargs": {},
            "path_semantics": {"posix": "posix", "nt": "windows"}[os.name],
        }

        self._resource_uid = new_uid()
        resource["uid"] = self._resource_uid
        self._asset_docs_cache.append(("resource", resource))

    def trigger(self):
        super().trigger()
        # save file stash file name
        self._result.clear()
        for idx, (name, reading) in enumerate(super().read().items()):
            # Save the actual reading['value'] to disk. For a real detector,
            # this part would be done by the detector IOC, not by ophyd.
            data_counter = next(self._datum_counter)
            self.save_func(
                "{}_{}.{}".format(self._path_stem, data_counter, self.save_ext),
                reading["value"],
            )
            # This is temporarily more complicated than it will be in the
            # future.  It needs to support old configurations that have a
            # registry.
            datum = {
                "resource": self._resource_uid,
                "datum_kwargs": dict(index=data_counter),
            }

            # If a Registry is not set, we need to generate the datum_id.
            datum_id = "{}/{}".format(self._resource_uid, data_counter)
            datum["datum_id"] = datum_id
            self._asset_docs_cache.append(("datum", datum))
            # And now change the reading in place, replacing the value with
            # a reference to Registry.
            reading["value"] = datum_id
            self._result[name] = reading

        return NullStatus()

    def read(self):
        return self._result

    def describe(self):
        res = super().describe()
        for key in res:
            res[key]["external"] = "FILESTORE"
        return res

    def collect_asset_docs(self):
        items = list(self._asset_docs_cache)
        self._asset_docs_cache.clear()
        for item in items:
            yield item

    def unstage(self):
        self._resource_uid = None
        self._datum_counter = None
        self._asset_docs_cache.clear()
        self._file_stem = None
        self._path_stem = None
        self._result.clear()


class NumpySeqHandler:
    specs = {"NPY_SEQ"}

    def __init__(self, filename, root=""):
        self._name = os.path.join(root, filename)

    def __call__(self, index):
        return np.load("{}_{}.npy".format(self._name, index), allow_pickle=False)

    def get_file_list(self, datum_kwarg_gen):
        "This method is optional. It is not needed for access, but for export."
        return [
            "{name}_{index}.npy".format(name=self._name, **kwargs)
            for kwargs in datum_kwarg_gen
        ]


class ABDetector(Device):
    a = Cpt(SynSignal, func=random.random, kind=Kind.hinted)
    b = Cpt(SynSignal, func=random.random)

    def trigger(self):
        return self.a.trigger() & self.b.trigger()


class DetWithCountTime(Device):
    intensity = Cpt(SynSignal, func=lambda: 0, kind=Kind.hinted)
    count_time = Cpt(Signal)


class DetWithConf(Device):
    a = Cpt(SynSignal, func=lambda: 1, kind=Kind.hinted)
    b = Cpt(SynSignal, func=lambda: 2, kind=Kind.hinted)
    c = Cpt(SynSignal, func=lambda: 3)
    d = Cpt(SynSignal, func=lambda: 4)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.read_attrs = ["a", "b"]
        self.configuration_attrs = ["c", "d"]

    def trigger(self):
        return self.a.trigger() & self.b.trigger()


class InvariantSignal(SynSignal):
    # Always returns the same reading, including timestamp.
    def read(self):
        res = super().read()
        for k in res:
            res[k]["timestamp"] = 0
        return res

    def __repr__(self):
        return "<INVARIANT REPR>"


class SPseudo3x3(PseudoPositioner):
    pseudo1 = Cpt(PseudoSingle, limits=(-10, 10), egu="a", kind=Kind.hinted)
    pseudo2 = Cpt(PseudoSingle, limits=(-10, 10), egu="b", kind=Kind.hinted)
    pseudo3 = Cpt(PseudoSingle, limits=None, egu="c", kind=Kind.hinted)
    real1 = Cpt(SoftPositioner, init_pos=0)
    real2 = Cpt(SoftPositioner, init_pos=0)
    real3 = Cpt(SoftPositioner, init_pos=0)

    sig = Cpt(Signal, value=0)

    @pseudo_position_argument
    def forward(self, pseudo_pos):
        pseudo_pos = self.PseudoPosition(*pseudo_pos)
        # logger.debug('forward %s', pseudo_pos)
        return self.RealPosition(
            real1=-pseudo_pos.pseudo1,
            real2=-pseudo_pos.pseudo2,
            real3=-pseudo_pos.pseudo3,
        )

    @real_position_argument
    def inverse(self, real_pos):
        real_pos = self.RealPosition(*real_pos)
        # logger.debug('inverse %s', real_pos)
        return self.PseudoPosition(
            pseudo1=-real_pos.real1, pseudo2=-real_pos.real2, pseudo3=-real_pos.real3
        )


class SPseudo1x3(PseudoPositioner):
    pseudo1 = Cpt(PseudoSingle, limits=(-10, 10), kind=Kind.hinted)
    real1 = Cpt(SoftPositioner, init_pos=0)
    real2 = Cpt(SoftPositioner, init_pos=0)
    real3 = Cpt(SoftPositioner, init_pos=0)

    @pseudo_position_argument
    def forward(self, pseudo_pos):
        pseudo_pos = self.PseudoPosition(*pseudo_pos)
        # logger.debug('forward %s', pseudo_pos)
        return self.RealPosition(
            real1=-pseudo_pos.pseudo1,
            real2=-pseudo_pos.pseudo1,
            real3=-pseudo_pos.pseudo1,
        )

    @real_position_argument
    def inverse(self, real_pos):
        real_pos = self.RealPosition(*real_pos)
        # logger.debug('inverse %s', real_pos)
        return self.PseudoPosition(pseudo1=-real_pos.real1)


class SynAxisNoPosition(SynAxis):
    @property
    def position(self):
        raise AttributeError


def make_fake_device(cls):
    """
    Inspect cls and construct a fake device that has the same structure.

    This works by replacing EpicsSignal with FakeEpicsSignal and EpicsSignalRO
    with FakeEpicsSignalRO. The fake class will be a subclass of the real
    class.

    This assumes that EPICS connections are done entirely in EpicsSignal and
    EpicsSignalRO subcomponents. If this is not true, this will fail silently
    on class construction and loudly when manipulating an object.

    Parameters
    ----------
    cls : Device
        A real Device class to inspect and create a fake Device class from

    Returns
    -------
    fake_device : Device
        The resulting fake Device class
    """
    # Cache to avoid repeating work.
    # EpicsSignal and EpicsSignalRO begin in the cache.
    if cls not in fake_device_cache:
        if not issubclass(cls, Device):
            # Ignore non-devices and non-epics-signals
            logger.debug("Ignore cls=%s, bases are %s", cls, cls.__bases__)
            fake_device_cache[cls] = cls
            return cls
        fake_dict = {}
        # Update all the components recursively
        for cpt_name in cls.component_names:
            cpt = getattr(cls, cpt_name)
            if isinstance(cpt, DDCpt):
                # Make a regular Cpt out of the DDC, as it already has
                # been generated
                fake_cpt = Cpt(
                    cpt.cls,
                    suffix=cpt.suffix,
                    lazy=cpt.lazy,
                    trigger_value=cpt.trigger_value,
                    kind=cpt.kind,
                    add_prefix=cpt.add_prefix,
                    doc=cpt.doc,
                    **cpt.kwargs,
                )
            else:
                fake_cpt = copy.copy(cpt)

            fake_cpt.cls = make_fake_device(cpt.cls)
            logger.debug("switch cpt_name=%s to cls=%s", cpt_name, fake_cpt.cls)

            fake_dict[cpt_name] = fake_cpt
        fake_class = type("Fake{}".format(cls.__name__), (cls,), fake_dict)
        fake_device_cache[cls] = fake_class
        logger.debug("fake_device_cache[%s] = %s", cls, fake_class)
    return fake_device_cache[cls]


def clear_fake_device(
    dev, *, default_value=0, default_string_value="", ignore_exceptions=False
):
    """Clear a fake device by setting all signals to a specific value

    Parameters
    ----------
    dev : Device
        The fake device
    default_value : any, optional
        The value to put to non-string components
    default_string_value : any, optional
        The value to put to components determined to be strings
    ignore_exceptions : bool, optional
        Ignore any exceptions raised by `sim_put`

    Returns
    -------
    all_values : list
        List of all (signal_instance, value) that were set
    """

    all_values = []
    for walk in dev.walk_signals(include_lazy=True):
        sig = walk.item
        if not hasattr(sig, "sim_put"):
            continue

        try:
            string = getattr(sig, "as_string", False)
            value = default_string_value if string else default_value
            sig.sim_put(value)
        except Exception:
            if not ignore_exceptions:
                raise
        else:
            all_values.append((sig, value))

    return all_values


def instantiate_fake_device(dev_cls, *, name=None, prefix="_prefix", **specified_kw):
    """Instantiate a fake device, optionally specifying some initializer kwargs

    If unspecified, all initializer keyword arguments will default to the
    string f"_{argument_name}_".

    Parameters
    ----------
    dev_cls : class
        The device class to instantiate. This is allowed to be a regular
        device, as `make_fake_device` will be called on it first.
    name : str, optional
        The instantiated device name
    prefix : str, optional
        The instantiated device prefix
    **specified_kw :
        Keyword arguments to override with a specific value

    Returns
    -------
    dev : dev_cls instance
        The instantiated fake device
    """
    dev_cls = make_fake_device(dev_cls)
    sig = inspect.signature(dev_cls)
    ignore_kw = {
        "kind",
        "read_attrs",
        "configuration_attrs",
        "parent",
        "args",
        "name",
        "prefix",
    }

    def get_kwarg(name, param):
        default = param.default
        if default == param.empty:
            # NOTE: could check param.annotation here
            default = "_{}_".format(param.name)
        return specified_kw.get(name, default)

    kwargs = {
        name: get_kwarg(name, param)
        for name, param in sig.parameters.items()
        if param.kind != param.VAR_KEYWORD and name not in ignore_kw
    }
    kwargs["name"] = name if name is not None else dev_cls.__name__
    kwargs["prefix"] = prefix
    return dev_cls(**kwargs)


class FakeEpicsSignal(SynSignal):
    """
    Fake version of EpicsSignal that's really just a SynSignal.

    Wheras SynSignal is generally used to test plans, FakeEpicsSignal is
    generally used in conjunction with make_fake_device to test any logic
    inside of a Device subclass.

    Unlike in SynSignal, this class is generally instantiated inside of a
    subcomponent generated automatically by make_fake_device. This means we
    need extra hooks for modifying the signal's properties after the class
    instantiates.

    We can emulate EpicsSignal features here. We currently emulate the put
    limits and some enum handling.
    """

    _metadata_keys = EpicsSignal._metadata_keys

    def __init__(
        self,
        read_pv,
        write_pv=None,
        *,
        put_complete=False,
        string=False,
        limits=False,
        auto_monitor=False,
        name=None,
        timeout=None,
        write_timeout=None,
        connection_timeout=None,
        **kwargs,
    ):
        """
        Mimic EpicsSignal signature
        """
        self.as_string = string
        self._enum_strs = None
        super().__init__(name=name, **kwargs)
        self._use_limits = limits
        self._put_func = None
        self._limits = None
        self._metadata.update(
            connected=True,
        )

    def describe(self):
        desc = super().describe()
        if self._enum_strs is not None:
            desc[self.name]["enum_strs"] = self.enum_strs
        return desc

    def sim_set_putter(self, putter):
        """
        Define arbirary behavior on signal put.

        This can be used to emulate basic IOC behavior.
        """
        self._put_func = putter

    def get(self, *, as_string=None, connection_timeout=1.0, **kwargs):
        """
        Implement getting as enum strings
        """
        if as_string is None:
            as_string = self.as_string

        value = super().get()

        if as_string:
            if self.enum_strs is not None and isinstance(value, int):
                return self.enum_strs[value]
            elif value is not None:
                return str(value)
        return value

    def put(
        self,
        value,
        *args,
        connection_timeout=0.0,
        callback=None,
        use_complete=None,
        timeout=0.0,
        wait=True,
        **kwargs,
    ):
        """
        Implement putting as enum strings and put functions

        Notes
        -----
        FakeEpicsSignal varies in subtle ways from the real class.

        * put-completion callback will _not_ be called.
        * connection_timeout, use_complete, wait, and timeout are ignored.
        """
        if self.enum_strs is not None:
            if value in self.enum_strs:
                value = self.enum_strs.index(value)
            elif isinstance(value, str):
                err = "{} not in enum strs {}".format(value, self.enum_strs)
                raise ValueError(err)
        if self._put_func is not None:
            return self._put_func(value, *args, **kwargs)
        return super().put(value, *args, **kwargs)

    def sim_put(self, *args, **kwargs):
        """
        Update the read-only signal's value.

        Implement here instead of FakeEpicsSignalRO so you can call it with
        every fake signal.
        """
        force = kwargs.pop("force", True)
        # The following will emit SUB_VALUE:
        ret = Signal.put(self, *args, force=force, **kwargs)
        # Also, ensure that SUB_META has been emitted:
        self._run_subs(sub_type=self.SUB_META, **self._metadata)
        return ret

    @property
    def enum_strs(self):
        """
        Simulated enum strings.

        Use sim_set_enum_strs during setup to set the enum strs.
        """
        return self._enum_strs

    def sim_set_enum_strs(self, enums):
        """
        Set the enum_strs for a fake device

        Parameters
        ----------
        enums: list or tuple of str
            The enums will be accessed by array index, e.g. the first item in
            enums will be 0, the next will be 1, etc.
        """
        self._enum_strs = tuple(enums)
        self._metadata["enum_strs"] = tuple(enums)
        self._run_subs(sub_type=self.SUB_META, **self._metadata)

    @property
    def limits(self):
        return self._limits

    def sim_set_limits(self, limits):
        """
        Set the fake signal's limits.
        """
        self._limits = limits

    def check_value(self, value):
        """
        Implement some of the checks from EpicsSignal
        """
        super().check_value(value)
        if value is None:
            raise ValueError("Cannot write None to EPICS PVs")
        if self._use_limits and not self.limits[0] <= value <= self.limits[1]:
            raise LimitError(f"value={value} not within limits {self.limits}")


class FakeEpicsSignalRO(SynSignalRO, FakeEpicsSignal):
    """
    Read-only FakeEpicsSignal
    """

    _metadata_keys = EpicsSignalRO._metadata_keys


class FakeEpicsSignalWithRBV(FakeEpicsSignal):
    """
    FakeEpicsSignal with PV and PV_RBV; used in the AreaDetector PV naming
    scheme
    """

    _metadata_keys = EpicsSignalWithRBV._metadata_keys

    def __init__(self, prefix, **kwargs):
        super().__init__(prefix + "_RBV", write_pv=prefix, **kwargs)


class FakeEpicsPathSignal(FakeEpicsSignal):
    """
    FakeEpicsPathSignal; used in AreaDetector for interacting with paths
    """

    _metadata_keys = EpicsPathSignal._metadata_keys

    def __init__(self, prefix, path_semantics, **kwargs):
        super().__init__(prefix + "_RBV", write_pv=prefix, **kwargs)


fake_device_cache = {
    EpicsSignal: FakeEpicsSignal,
    EpicsSignalRO: FakeEpicsSignalRO,
    EpicsSignalWithRBV: FakeEpicsSignalWithRBV,
    EpicsPathSignal: FakeEpicsPathSignal,
}


class DirectImage(Device):
    img = Cpt(SynSignal, kind="hinted")

    def __init__(self, *args, func=None, **kwargs):
        super().__init__(*args, **kwargs)
        if func is not None:
            self.img.sim_set_func(func)

    def trigger(self):
        return self.img.trigger()


def hw(save_path=None):
    "Build a set of synthetic hardware (hence the abbreviated name, hw)"
    motor = SynAxis(name="motor", labels={"motors"})
    motor1 = SynAxis(name="motor1", labels={"motors"})
    motor2 = SynAxis(name="motor2", labels={"motors"})
    motor3 = SynAxis(name="motor3", labels={"motors"})
    jittery_motor1 = SynAxis(
        name="jittery_motor1",
        readback_func=lambda x: x + np.random.rand(),
        labels={"motors"},
    )
    jittery_motor2 = SynAxis(
        name="jittery_motor2",
        readback_func=lambda x: x + np.random.rand(),
        labels={"motors"},
    )
    noisy_det = SynGauss(
        "noisy_det",
        motor,
        "motor",
        center=0,
        Imax=1,
        noise="uniform",
        sigma=1,
        noise_multiplier=0.1,
        labels={"detectors"},
    )
    det = SynGauss(
        "det", motor, "motor", center=0, Imax=1, sigma=1, labels={"detectors"}
    )
    identical_det = SynGauss(
        "det", motor, "motor", center=0, Imax=1, sigma=1, labels={"detectors"}
    )
    det1 = SynGauss(
        "det1", motor1, "motor1", center=0, Imax=5, sigma=0.5, labels={"detectors"}
    )
    det2 = SynGauss(
        "det2", motor2, "motor2", center=1, Imax=2, sigma=2, labels={"detectors"}
    )
    det3 = SynGauss(
        "det3", motor3, "motor3", center=-1, Imax=2, sigma=1, labels={"detectors"}
    )
    det4 = Syn2DGauss(
        "det4",
        motor1,
        "motor1",
        motor2,
        "motor2",
        center=(0, 0),
        Imax=1,
        labels={"detectors"},
    )
    det5 = Syn2DGauss(
        "det5",
        jittery_motor1,
        "jittery_motor1",
        jittery_motor2,
        "jittery_motor2",
        center=(0, 0),
        Imax=1,
        labels={"detectors"},
    )

    flyer1 = MockFlyer("flyer1", det, motor, 1, 5, 20)
    flyer2 = MockFlyer("flyer2", det, motor, 1, 5, 10)
    trivial_flyer = TrivialFlyer()
    new_trivial_flyer = NewTrivialFlyer()

    ab_det = ABDetector(name="det", labels={"detectors"})
    # area detector that directly stores image data in Event
    direct_img = DirectImage(
        func=lambda: np.array(np.ones((10, 10))), name="direct", labels={"detectors"}
    )
    direct_img.img.name = "img"

    direct_img_list = DirectImage(
        func=lambda: [[1] * 10] * 10, name="direct", labels={"detectors"}
    )
    direct_img_list.img.name = "direct_img_list"

    # area detector that stores data in a file
    img = SynSignalWithRegistry(
        func=lambda: np.array(np.ones((10, 10))),
        name="img",
        labels={"detectors"},
        save_path=save_path,
    )
    invariant1 = InvariantSignal(
        func=lambda: 0, name="invariant1", labels={"detectors"}
    )
    invariant2 = InvariantSignal(
        func=lambda: 0, name="invariant2", labels={"detectors"}
    )
    det_with_conf = DetWithConf(name="det", labels={"detectors"})
    det_with_count_time = DetWithCountTime(name="det", labels={"detectors"})
    rand = SynPeriodicSignal(name="rand", labels={"detectors"})
    rand2 = SynPeriodicSignal(name="rand2", labels={"detectors"})
    motor_no_pos = SynAxisNoPosition(name="motor", labels={"motors"})
    bool_sig = Signal(value=False, name="bool_sig", labels={"detectors"})

    motor_empty_hints1 = SynAxisEmptyHints(name="motor1", labels={"motors"})
    motor_empty_hints2 = SynAxisEmptyHints(name="motor2", labels={"motors"})

    motor_no_hints1 = SynAxisNoHints(name="motor1", labels={"motors"})
    motor_no_hints2 = SynAxisNoHints(name="motor2", labels={"motors"})
    # Because some of these reference one another we must define them (above)
    # before we pack them into a namespace (below).

    signal = SynSignal(name="signal")

    return SimpleNamespace(
        motor=motor,
        motor1=motor1,
        motor2=motor2,
        motor3=motor3,
        jittery_motor1=jittery_motor1,
        jittery_motor2=jittery_motor2,
        noisy_det=noisy_det,
        det=det,
        identical_det=identical_det,
        det1=det1,
        det2=det2,
        det3=det3,
        det4=det4,
        det5=det5,
        flyer1=flyer1,
        flyer2=flyer2,
        trivial_flyer=trivial_flyer,
        new_trivial_flyer=new_trivial_flyer,
        ab_det=ab_det,
        direct_img=direct_img,
        direct_img_list=direct_img_list,
        img=img,
        invariant1=invariant1,
        invariant2=invariant2,
        pseudo3x3=SPseudo3x3(name="pseudo3x3"),
        pseudo1x3=SPseudo1x3(name="pseudo1x3"),
        sig=Signal(name="sig", value=0),
        det_with_conf=det_with_conf,
        det_with_count_time=det_with_count_time,
        rand=rand,
        rand2=rand2,
        motor_no_pos=motor_no_pos,
        motor_empty_hints1=motor_empty_hints1,
        motor_empty_hints2=motor_empty_hints2,
        motor_no_hints1=motor_no_hints1,
        motor_no_hints2=motor_no_hints2,
        bool_sig=bool_sig,
        signal=signal,
    )


# Dump instances of the example hardware generated by hw() into the global
# namespcae for convenience and back-compat.
globals().update(hw().__dict__)



# vi: ts=4 sw=4
"""AreaDetector Devices

`areaDetector`_ detector abstractions

.. _areaDetector: https://areadetector.github.io/master/index.html
"""
import warnings

from . import cam
from .base import ADBase
from .base import ADComponent as C

__all__ = [
    "DetectorBase",
    "AreaDetector",
    "AdscDetector",
    "Andor3Detector",
    "AndorDetector",
    "BrukerDetector",
    "DexelaDetector",
    "EmergentVisionDetector",
    "EigerDetector",
    "FirewireLinDetector",
    "FirewireWinDetector",
    "GreatEyesDetector",
    "LightFieldDetector",
    "Mar345Detector",
    "MarCCDDetector",
    "PSLDetector",
    "PerkinElmerDetector",
    "PICamDetector",
    "PilatusDetector",
    "PixiradDetector",
    "PointGreyDetector",
    "ProsilicaDetector",
    "PvcamDetector",
    "RoperDetector",
    "SimDetector",
    "URLDetector",
    "UVCDetector",
    "Xspress3Detector",
]


class DetectorBase(ADBase):
    """
    The base class for the hardware-specific classes that follow.

    Note that Plugin also inherits from ADBase.
    This adds some AD-specific methods that are not shared by the plugins.
    """

    _default_configuration_attrs = ADBase._default_configuration_attrs + ("cam",)

    def generate_datum(self, key, timestamp, datum_kwargs=None):
        """
        Notify plugins of acquisition being complete.

        When a new acquisition is started, this method is called with a
        key which is a label like 'light', 'dark', or 'gain8'.

        It in turn calls ``generate_datum`` on all of the plugins that have
        that method.

        File plugins are identified by searching for a
        :meth:`~ophyd.areadetector.filestore_mixins.FileStoreBase.generate_datum`
        method that must have the signature ::

           def generate_datum(key: str, timestamp: float, datum_kwargs: dict):
              ...

        Parameters
        ----------
        key : str
            The label for the datum that should be generated

        timestamp : float
            The time of the trigger

        datum_kwargs : Dict[str, Any], optional
            Any datum kwargs that should go to all children.
        """
        if datum_kwargs is None:
            datum_kwargs = {}
        file_plugins = [
            s for s in self._signals.values() if hasattr(s, "generate_datum")
        ]
        for p in file_plugins:
            if p.enable.get():
                p.generate_datum(key, timestamp, datum_kwargs)

    def dispatch(self, key, timestamp):
        warnings.warn(
            ".dispatch is deprecated, use .generate_datum instead", stacklevel=2
        )

        return self.generate_datum(key, timestamp, {})

    dispatch.__doc__ = generate_datum.__doc__

    def make_data_key(self):
        source = "PV:{}".format(self.prefix)
        # This shape is expected to match arr.shape for the array.
        shape = (
            self.cam.num_images.get(),
            self.cam.array_size.array_size_y.get(),
            self.cam.array_size.array_size_x.get(),
        )
        return dict(shape=shape, source=source, dtype="array", external="FILESTORE:")

    def collect_asset_docs(self):
        file_plugins = [
            s for s in self._signals.values() if hasattr(s, "collect_asset_docs")
        ]
        for p in file_plugins:
            yield from p.collect_asset_docs()


class AreaDetector(DetectorBase):
    cam = C(cam.AreaDetectorCam, "cam1:")


class SimDetector(DetectorBase):
    _html_docs = ["simDetectorDoc.html"]
    cam = C(cam.SimDetectorCam, "cam1:")


class AdscDetector(DetectorBase):
    _html_docs = ["adscDoc.html"]
    cam = C(cam.AdscDetectorCam, "cam1:")


class AndorDetector(DetectorBase):
    _html_docs = ["andorDoc.html"]
    cam = C(cam.AndorDetectorCam, "cam1:")


class Andor3Detector(DetectorBase):
    _html_docs = ["andor3Doc.html"]
    cam = C(cam.Andor3DetectorCam, "cam1:")


class BrukerDetector(DetectorBase):
    _html_docs = ["BrukerDoc.html"]
    cam = C(cam.BrukerDetectorCam, "cam1:")


class DexelaDetector(DetectorBase):
    _html_docs = ["DexelaDoc.html"]
    cam = C(cam.DexelaDetectorCam, "cam1:")


class EmergentVisionDetector(DetectorBase):
    _html_docs = ["EVTDoc.html"]
    cam = C(cam.EmergentVisionDetectorCam, "cam1:")


class EigerDetector(DetectorBase):
    _html_docs = ["EigerDoc.html"]
    cam = C(cam.EigerDetectorCam, "cam1:")


class FirewireLinDetector(DetectorBase):
    _html_docs = ["FirewireWinDoc.html"]
    cam = C(cam.FirewireLinDetectorCam, "cam1:")


class FirewireWinDetector(DetectorBase):
    _html_docs = ["FirewireWinDoc.html"]
    cam = C(cam.FirewireWinDetectorCam, "cam1:")


class GreatEyesDetector(DetectorBase):
    _html_docs = []  # the documentation is not public
    cam = C(cam.GreatEyesDetectorCam, "cam1:")


class LightFieldDetector(DetectorBase):
    _html_docs = ["LightFieldDoc.html"]
    cam = C(cam.LightFieldDetectorCam, "cam1:")


class Mar345Detector(DetectorBase):
    _html_docs = ["Mar345Doc.html"]
    cam = C(cam.Mar345DetectorCam, "cam1:")


class MarCCDDetector(DetectorBase):
    _html_docs = ["MarCCDDoc.html"]
    cam = C(cam.MarCCDDetectorCam, "cam1:")


class PerkinElmerDetector(DetectorBase):
    _html_docs = ["PerkinElmerDoc.html"]
    cam = C(cam.PerkinElmerDetectorCam, "cam1:")


class PSLDetector(DetectorBase):
    _html_docs = ["PSLDoc.html"]
    cam = C(cam.PSLDetectorCam, "cam1:")


class PICamDetector(DetectorBase):
    _html_docs = ["PICamDoc.html"]
    cam = C(cam.PICamDetectorCam, "cam1:")


class PilatusDetector(DetectorBase):
    _html_docs = ["pilatusDoc.html"]
    cam = C(cam.PilatusDetectorCam, "cam1:")


class PixiradDetector(DetectorBase):
    _html_docs = ["PixiradDoc.html"]
    cam = C(cam.PixiradDetectorCam, "cam1:")


class PointGreyDetector(DetectorBase):
    _html_docs = ["PointGreyDoc.html"]
    cam = C(cam.PointGreyDetectorCam, "cam1:")


class ProsilicaDetector(DetectorBase):
    _html_docs = ["prosilicaDoc.html"]
    cam = C(cam.ProsilicaDetectorCam, "cam1:")


class PvcamDetector(DetectorBase):
    _html_docs = ["pvcamDoc.html"]
    cam = C(cam.PvcamDetectorCam, "cam1:")


class RoperDetector(DetectorBase):
    _html_docs = ["RoperDoc.html"]
    cam = C(cam.RoperDetectorCam, "cam1:")


class URLDetector(DetectorBase):
    _html_docs = ["URLDoc.html"]
    cam = C(cam.URLDetectorCam, "cam1:")


class UVCDetector(DetectorBase):
    _html_docs = ["UVCDoc.html"]
    cam = C(cam.UVCDetectorCam, "cam1:")


class Xspress3Detector(DetectorBase):
    _html_docs = ["Xspress3Doc.html"]
    cam = C(cam.Xspress3DetectorCam, "det1:")



import time
import logging
import pytest
from copy import copy
from numpy.testing import assert_allclose

from ophyd import (
    EpicsMotor,
    Component as Cpt,
    MotorBundle,
)
from ophyd.utils.epics_pvs import AlarmSeverity, AlarmStatus
from ophyd.utils import UnknownStatusFailure
import threading

logger = logging.getLogger(__name__)


@pytest.mark.motorsim
def test_timeout(motor):
    assert motor.timeout == 10.0
    motor.timeout = 20.0
    assert motor.timeout == 20.0


@pytest.mark.motorsim
def test_connected(motor):
    assert motor.connected


@pytest.mark.motorsim
def test_limits(motor):
    device_limits = (motor.low_limit_value.get(), motor.high_limit_value.get())
    assert motor.limits == device_limits


@pytest.mark.motorsim
def test_checkvalue(motor):
    motor.check_value(0)


@pytest.mark.motorsim
def test_move(motor):
    motor.stop()
    motor.move(0.1, timeout=5, wait=True)

    logger.debug("Move to 0.0")
    motor.move(0.0, timeout=5, wait=True)
    time.sleep(0.1)
    assert_allclose(motor.position, 0.0)

    assert motor.settle_time == 0.1

    logger.debug("Move to 0.1")
    motor.move(0.1, timeout=5, wait=True)
    time.sleep(0.1)
    assert_allclose(motor.position, 0.1)

    logger.debug("Move to 0.1, again")
    motor.move(0.1, timeout=5, wait=True)
    time.sleep(0.1)
    assert_allclose(motor.position, 0.1)

    logger.debug("Move to 0.0")
    motor.move(0.0, timeout=5, wait=True)
    time.sleep(0.1)
    assert_allclose(motor.position, 0.0)


@pytest.mark.motorsim
def test_copy(motor):
    repr(motor)
    str(motor)

    mc = copy(motor)
    assert mc.prefix == motor.prefix

    res = motor.move(0.2, wait=False)

    while not res.done:
        time.sleep(0.1)

    time.sleep(0.1)
    assert res.settle_time == 0.1
    assert_allclose(motor.position, 0.2)

    motor.settle_time = 0.2
    assert motor.settle_time == 0.2

    assert res.done
    assert_allclose(res.error, 0, atol=1e-6)
    assert res.elapsed > 0


@pytest.mark.motorsim
def test_read(motor):
    motor.read()


@pytest.mark.motorsim
def test_report(motor):
    motor.report


@pytest.mark.motorsim
def test_calibration(motor):
    motor.user_offset.put(0, wait=True)

    # Calibration
    old_position = motor.position
    expected_offset = 10 - motor.position
    motor.set_current_position(10)
    assert motor.offset_freeze_switch.get() == 0
    assert_allclose(motor.position, 10)
    assert motor.user_offset.get() == expected_offset
    motor.set_current_position(old_position)


@pytest.mark.motorsim
def test_high_limit_switch(motor):
    # limit switch status
    motor.direction_of_travel.put(1)
    motor.high_limit_switch.put(1)
    res = motor.move(1, wait=False)
    repr(res)
    assert res.timeout == 10.0

    with pytest.raises(UnknownStatusFailure):
        res.wait(11)
    assert motor.high_limit_switch.get() == 1

    motor.high_limit_switch.put(0)


@pytest.mark.motorsim
def test_low_limit_switch(motor):
    motor.direction_of_travel.put(0)
    motor.low_limit_switch.put(1)
    res = motor.move(0, wait=False)

    with pytest.raises(UnknownStatusFailure):
        res.wait(11)
    assert motor.low_limit_switch.get() == 1
    motor.low_limit_switch.put(0)


@pytest.mark.motorsim
def test_low_limit_switch_while_moving_out(motor):
    # If the Motor is at the Low Limit Switch
    # all the movements in the opposite
    # direction must be success=True
    motor.direction_of_travel.put(1)
    motor.low_limit_switch.put(1)
    res = motor.move(1, wait=False)

    while not res.done:
        time.sleep(0.1)

    assert motor.low_limit_switch.get() == 1
    assert res.success
    motor.low_limit_switch.put(0)


@pytest.mark.motorsim
def test_high_limit_switch_while_moving_out(motor):
    # If the Motor is at the High Limit Switch
    # all the movements in the opposite
    # direction must be success=True
    motor.direction_of_travel.put(0)
    motor.high_limit_switch.put(1)
    res = motor.move(0, wait=False)

    while not res.done:
        time.sleep(0.1)

    assert motor.high_limit_switch.get() == 1
    assert res.success
    motor.high_limit_switch.put(0)


# @pytest.mark.skip(reason="This has become flaky, not sure why")
@pytest.mark.motorsim
def test_homing_forward(motor):
    # homing forward
    motor.wait_for_connection(all_signals=True, timeout=2)
    motor.move(-1, wait=True)
    res = motor.home("forward", timeout=2, wait=False)

    while not res.done:
        time.sleep(0.1)

    # MotorSim is unable to execute homing
    assert not res.success
    motor.stop()


# @pytest.mark.skip(reason="This has become flaky, not sure why")
@pytest.mark.motorsim
def test_homing_reverse(motor):
    # homing reverse
    motor.move(1, wait=True)
    res = motor.home("reverse", timeout=2, wait=False)

    while not res.done:
        time.sleep(0.1)

    assert not res.success
    motor.stop()


@pytest.mark.motorsim
def test_homing_invalid(motor):
    with pytest.raises(ValueError):
        # homing reverse
        motor.move(1, wait=True)
        res = motor.home("foobar", timeout=2, wait=False)

        while not res.done:
            time.sleep(0.1)

        assert not res.success
        motor.stop()


@pytest.mark.motorsim
def test_move_alarm(motor):
    try:
        motor.user_readback.alarm_status = AlarmStatus.COMM
        motor.user_readback.alarm_severity = AlarmSeverity.MAJOR

        if motor.position + 1 < motor.high_limit_value.get():
            target_pos = motor.position + 1
        else:
            target_pos = motor.position - 1

        st = motor.move(target_pos, wait=False)

        while not st.done:
            time.sleep(0.1)

        assert not st.success
    finally:
        motor.user_readback.alarm_status = AlarmStatus.NO_ALARM
        motor.user_readback.alarm_severity = AlarmSeverity.NO_ALARM


@pytest.mark.motorsim
def test_hints(motor):
    assert motor.hints == {"fields": list(motor.user_readback.read())}

    motor.user_setpoint.kind = "hinted"
    motor.user_readback.kind = "normal"
    assert motor.hints == {"fields": list(motor.user_setpoint.read())}


@pytest.mark.motorsim
def test_watchers(motor):

    st = motor.set(0)
    while not st.done:
        continue

    collector = []

    def collect(fraction, **kwargs):
        collector.append(fraction)

    def callback(status):
        ev.set()

    st = motor.set(1)
    st.watch(collect)
    ev = threading.Event()
    st.add_callback(callback)
    ev.wait()
    assert collector
    assert collector[-1] == 0.0
    assert len(collector) > 1


@pytest.mark.motorsim
def test_str_smoke(motor):
    str(motor)


@pytest.mark.motorsim
def test_read_in_motor_callback(motor):
    out_cache = []
    motor.set(0, wait=True)

    def cb(**kwargs):
        v = motor.read()
        out_cache.append(v)

    motor.subscribe(cb)
    motor.set(0.1, wait=True)
    assert len(out_cache) > 0


def test_motor_bundle():
    class Bundle(MotorBundle):
        a = Cpt(EpicsMotor, ":mtr1")
        b = Cpt(EpicsMotor, ":mtr2")
        c = Cpt(EpicsMotor, ":mtr3")

    bundle = Bundle("sim", name="bundle")

    assert bundle.hints["fields"] == ["bundle_{}".format(k) for k in "abc"]

    # Test old-style attributes.
    assert set(bundle.read_attrs) == set(
        list("abc")
        + [".".join([p, c]) for p in "abc" for c in ["user_readback", "user_setpoint"]]
    )
    assert set(bundle.configuration_attrs) == set(
        list("abc")
        + [
            ".".join([p, c])
            for p in "abc"
            for c in [
                "user_offset",
                "user_offset_dir",
                "velocity",
                "acceleration",
                "motor_egu",
            ]
        ]
    )
